# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
SPI Flash Descriptor binary parsing functionality


usage:
    >>> fd = read_file( fd_file )
    >>> parse_spi_flash_descriptor( fd )
"""

import struct
from typing import Dict, List, Optional, Tuple
from chipsec.library.logger import logger, print_buffer_bytes
from chipsec.hal import spi

SPI_FLASH_DESCRIPTOR_SIGNATURE = struct.pack('=I', 0x0FF0A55A)
SPI_FLASH_DESCRIPTOR_SIZE = 0x1000


def get_spi_flash_descriptor(rom: bytes) -> Tuple[int, bytes]:
    pos = rom.find(SPI_FLASH_DESCRIPTOR_SIGNATURE)
    if (-1 == pos or pos < 0x10):
        return (-1, b'')
    fd_off = pos - 0x10
    fd = rom[fd_off: fd_off + SPI_FLASH_DESCRIPTOR_SIZE]
    return (fd_off, fd)


def get_SPI_master(flmstr: int) -> Tuple[int, int, int]:
    requester_id = (flmstr & 0xFFFF)
    master_region_ra = ((flmstr >> 16) & 0xFF)
    master_region_wa = ((flmstr >> 24) & 0xFF)
    return (requester_id, master_region_ra, master_region_wa)


def get_spi_regions(fd: bytes) -> Optional[List[Tuple[int, str, int, int, int, bool]]]:
    pos = fd.find(SPI_FLASH_DESCRIPTOR_SIGNATURE)
    if not (pos == 0x10):
        return None

    flmap0 = struct.unpack_from('=I', fd[0x14:0x18])[0]
    # Flash Region Base Address (bits [23:16])
    frba = ((flmap0 & 0x00FF0000) >> 12)

    flregs = []
    for r in range(spi.SPI_REGION_NUMBER_IN_FD):
        flreg_off = frba + r * 4
        flreg = struct.unpack_from('=I', fd[flreg_off:flreg_off + 0x4])[0]
        (base, limit) = spi.get_SPI_region(flreg)
        notused = (base > limit)
        flregs.append((r, spi.SPI_REGION_NAMES[r], flreg, base, limit, notused))

    fd_size = flregs[spi.FLASH_DESCRIPTOR][4] - flregs[spi.FLASH_DESCRIPTOR][3] + 1
    fd_notused = flregs[spi.FLASH_DESCRIPTOR][5]
    if fd_notused or (fd_size != SPI_FLASH_DESCRIPTOR_SIZE):
        return None

    return flregs


def parse_spi_flash_descriptor(cs, rom: bytes) -> None:
    if not (isinstance(rom, str) or isinstance(rom, bytes)):
        logger().log_error(f'Invalid fd object type {type(rom)}')
        return

    pos = rom.find(SPI_FLASH_DESCRIPTOR_SIGNATURE)
    if (-1 == pos) or (pos < 0x10):
        desc_signature = struct.unpack('=I', SPI_FLASH_DESCRIPTOR_SIGNATURE)[0]
        logger().log_error(f'Valid SPI flash descriptor is not found (should have signature {desc_signature:08X})')
        return None

    fd_off = pos - 0x10
    logger().log(f'[spi_fd] Valid SPI flash descriptor found at offset 0x{fd_off:08X}')

    logger().log('')
    logger().log('########################################################')
    logger().log('# SPI FLASH DESCRIPTOR')
    logger().log('########################################################')
    logger().log('')

    fd = rom[fd_off: fd_off + SPI_FLASH_DESCRIPTOR_SIZE]
    fd_sig = struct.unpack_from('=I', fd[0x10:0x14])[0]

    logger().log(f'+ 0x0000 Reserved : 0x{fd[0x0:0xF].hex().upper()}')
    logger().log(f'+ 0x0010 Signature: 0x{fd_sig:08X}')

    #
    # Flash Descriptor Map Section
    #
    flmap0 = struct.unpack_from('=I', fd[0x14:0x18])[0]
    flmap1 = struct.unpack_from('=I', fd[0x18:0x1C])[0]
    flmap2 = struct.unpack_from('=I', fd[0x1C:0x20])[0]
    cs.register.print('FLMAP0', flmap0)
    cs.register.print('FLMAP1', flmap1)
    cs.register.print('FLMAP2', flmap2)

    fcba = cs.register.get_field('FLMAP0', flmap0, 'FCBA')
    nc = cs.register.get_field('FLMAP0', flmap0, 'NC')
    frba = cs.register.get_field('FLMAP0', flmap0, 'FRBA')
    fcba = fcba << 4
    frba = frba << 4
    nc += 1
    logger().log('')
    logger().log('+ 0x0014 Flash Descriptor Map:')
    logger().log('========================================================')
    logger().log(f'  Flash Component Base Address: 0x{fcba:08X}')
    logger().log(f'  Flash Region Base Address   : 0x{frba:08X}')
    logger().log(f'  Number of Flash Components  : {nc:d}')

    nr = spi.SPI_REGION_NUMBER_IN_FD
    if cs.register.has_field('FLMAP0', 'NR'):
        nr = cs.register.get_field('FLMAP0', flmap0, 'NR')
        if nr == 0:
            logger().log_warning('only 1 region (FD) is found. Looks like flash descriptor binary is from Skylake platform or later. Try with option --platform')
        nr += 1
        logger().log(f'  Number of Regions           : {nr:d}')

    fmba = cs.register.get_field('FLMAP1', flmap1, 'FMBA')
    nm = cs.register.get_field('FLMAP1', flmap1, 'NM')
    fpsba = cs.register.get_field('FLMAP1', flmap1, 'FPSBA')
    psl = cs.register.get_field('FLMAP1', flmap1, 'PSL')
    fmba = fmba << 4
    fpsba = fpsba << 4
    logger().log(f'  Flash Master Base Address   : 0x{fmba:08X}')
    logger().log(f'  Number of Masters           : {nm:d}')
    logger().log(f'  Flash PCH Strap Base Address: 0x{fpsba:08X}')
    logger().log(f'  PCH Strap Length            : 0x{psl:X}')

    fcpusba = cs.register.get_field('FLMAP2', flmap2, 'FCPUSBA')
    cpusl = cs.register.get_field('FLMAP2', flmap2, 'CPUSL')
    logger().log(f'  Flash CPU Strap Base Address: 0x{fcpusba:08X}')
    logger().log(f'  CPU Strap Length            : 0x{cpusl:X}')

    #
    # Flash Descriptor Component Section
    #
    logger().log('')
    logger().log(f'+ 0x{fcba:04X} Component Section:')
    logger().log('========================================================')

    flcomp = struct.unpack_from('=I', fd[fcba + 0x0:fcba + 0x4])[0]
    logger().log(f'+ 0x{fcba:04X} FLCOMP   : 0x{flcomp:08X}')
    flil = struct.unpack_from('=I', fd[fcba + 0x4:fcba + 0x8])[0]
    logger().log(f'+ 0x{fcba + 0x4:04X} FLIL     : 0x{flil:08X}')
    flpb = struct.unpack_from('=I', fd[fcba + 0x8:fcba + 0xC])[0]
    logger().log(f'+ 0x{fcba + 0x8:04X} FLPB     : 0x{flpb:08X}')

    #
    # Flash Descriptor Region Section
    #
    logger().log('')
    logger().log(f'+ 0x{frba:04X} Region Section:')
    logger().log('========================================================')

    flregs: Dict[int, Tuple[int, int, int, str]] = {}
    for r in range(nr):
        flreg_off = frba + r * 4
        flreg = struct.unpack_from('=I', fd[flreg_off:flreg_off + 0x4])[0]
        if not cs.register.is_defined(f'FLREG{r:d}'):
            continue
        base = cs.register.get_field((f'FLREG{r:d}'), flreg, 'RB') << spi.SPI_FLA_SHIFT
        limit = cs.register.get_field((f'FLREG{r:d}'), flreg, 'RL') << spi.SPI_FLA_SHIFT
        notused = '(not used)' if base > limit or flreg == 0xFFFFFFFF else ''
        flregs[r] = (flreg, base, limit, notused)
        logger().log(f'+ 0x{flreg_off:04X} FLREG{r:d}   : 0x{flreg:08X} {notused}')

    logger().log('')
    logger().log('Flash Regions')
    logger().log('--------------------------------------------------------')
    logger().log(' Region                | FLREGx    | Base     | Limit   ')
    logger().log('--------------------------------------------------------')
    for r in flregs:
        if flregs[r]:
            logger().log(f'{r:d} {spi.SPI_REGION_NAMES[r]:20s} | {flregs[r][0]:08X}  | {flregs[r][1]:08X} | {flregs[r][2]:08X} {flregs[r][3]}')

    #
    # Flash Descriptor Master Section
    #
    logger().log('')
    logger().log(f'+ 0x{fmba:04X} Master Section:')
    logger().log('========================================================')

    flmstrs: Dict[int, Tuple[int, int]] = {}
    for m in range(nm):
        flmstr_off = fmba + m * 4
        flmstr = struct.unpack_from('=I', fd[flmstr_off:flmstr_off + 0x4])[0]
        master_region_ra = cs.register.get_field('FLMSTR1', flmstr, 'MRRA')
        master_region_wa = cs.register.get_field('FLMSTR1', flmstr, 'MRWA')
        flmstrs[m] = (master_region_ra, master_region_wa)
        logger().log(f'+ 0x{flmstr_off:04X} FLMSTR{m + 1:d}   : 0x{flmstr:08X}')

    logger().log('')
    logger().log('Master Read/Write Access to Flash Regions')
    logger().log('--------------------------------------------------------')
    s = ' Region                 '
    for m in range(nm):
        if m in spi.SPI_MASTER_NAMES:
            s = f'{s}| {spi.SPI_MASTER_NAMES[m]:9}'
        else:
            s = f'{s}| Master {m:-2d}'
    logger().log(s)
    logger().log('--------------------------------------------------------')
    for r in range(nr):
        s = f'{r:-2d} {spi.SPI_REGION_NAMES[r]:20s} '
        for m in range(nm):
            access_s = ''
            mask = (0x1 << r)
            if (flmstrs[m][0] & mask):
                access_s += 'R'
            if (flmstrs[m][1] & mask):
                access_s += 'W'
            s = f'{s}| {access_s:9}'
        logger().log(s)

    #
    # Flash Descriptor Upper Map Section
    #
    logger().log('')
    logger().log(f'+ 0x{0xEFC:04X} Flash Descriptor Upper Map:')
    logger().log('========================================================')

    flumap1 = struct.unpack_from('=I', fd[0xEFC:0xF00])[0]
    logger().log(f'+ 0x{0xEFC:04X} FLUMAP1   : 0x{flumap1:08X}')

    vtba = ((flumap1 & 0x000000FF) << 4)
    vtl = (((flumap1 & 0x0000FF00) >> 8) & 0xFF)
    logger().log(f'  VSCC Table Base Address    = 0x{vtba:08X}')
    logger().log(f'  VSCC Table Length          = 0x{vtl:02X}')

    #
    # OEM Section
    #
    logger().log('')
    logger().log(f'+ 0x{0xF00:04X} OEM Section:')
    logger().log('========================================================')
    print_buffer_bytes(fd[0xF00:])

    logger().log('')
    logger().log('########################################################')
    logger().log('# END OF SPI FLASH DESCRIPTOR')
    logger().log('########################################################')



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2025, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import unittest

from unittest.mock import MagicMock, patch
from chipsec.hal.smbus import SMBus, SMBUS_POLL_COUNT
from chipsec.library.exceptions import IOBARNotFoundError, RegisterNotFoundError


class TestSMBUS(unittest.TestCase):
    @patch("chipsec.hal.smbus.iobar")
    def test_get_SMBus_Base_Address_valid_base(self, mock_iobar):
        base_address = 123456
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        smbus_hal.iobar.is_IO_BAR_defined.return_value = True
        smbus_hal.iobar.get_IO_BAR_base_address.return_value = (base_address, None)
        self.assertEqual(smbus_hal.get_SMBus_Base_Address(), base_address)

    @patch("chipsec.hal.smbus.iobar")
    def test_get_SMBus_Base_Address_invalid_base(self, mock_iobar):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        smbus_hal.iobar.is_IO_BAR_defined.return_value = False
        self.assertRaises(IOBARNotFoundError, smbus_hal.get_SMBus_Base_Address)

    def test_get_SMBus_HCFG_valid(self):
        reg_read_val = 123456
        mock_cs = MagicMock()
        mock_cs.register.is_defined.return_value = True
        mock_cs.register.read.return_value = reg_read_val
        smbus_hal = SMBus(mock_cs)
        self.assertEqual(smbus_hal.get_SMBus_HCFG(), reg_read_val)

    def test_get_SMBus_HCFG_invalid(self):
        mock_cs = MagicMock()
        mock_cs.register.is_defined.return_value = False
        smbus_hal = SMBus(mock_cs)
        self.assertRaises(RegisterNotFoundError, smbus_hal.get_SMBus_HCFG)

    @patch("chipsec.hal.smbus.iobar")
    def test_display_SMBus_info(self, mock_iobar):
        reg_read_val = 123456
        base_address = 456789
        mock_cs = MagicMock()
        mock_cs.register.is_defined.return_value = True
        mock_cs.register.read.return_value = reg_read_val
        smbus_hal = SMBus(mock_cs)
        smbus_hal.logger = MagicMock()
        smbus_hal.logger.HAL = True
        smbus_hal.iobar.is_IO_BAR_defined.return_value = True
        smbus_hal.iobar.get_IO_BAR_base_address.return_value = (base_address, None)
        smbus_hal.display_SMBus_info()
        smbus_hal.logger.log_hal.assert_called_with(f'[smbus] SMBus Base Address: 0x{base_address:04X}')

    def test_is_SMBus_enabled(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.device.is_enabled.return_value = True
        self.assertTrue(smbus_hal.is_SMBus_enabled())

    def test_is_SMBus_supported_valid(self):
        did = 1234
        vid = 0x8086
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.device.get_VendorID.return_value = (did, vid)
        self.assertTrue(smbus_hal.is_SMBus_supported())

    def test_is_SMBus_supported_invalid(self):
        did = 1234
        vid = 0x5678
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.device.get_VendorID.return_value = (did, vid)
        self.assertFalse(smbus_hal.is_SMBus_supported())

    def test_is_SMBus_host_controller_enabled_valid(self):
        reg_read_val = 123456
        hst_en = 1
        mock_cs = MagicMock()
        mock_cs.register.is_defined.return_value = True
        mock_cs.register.read.return_value = reg_read_val
        mock_cs.register.get_field.return_value = hst_en
        smbus_hal = SMBus(mock_cs)
        self.assertEqual(smbus_hal.is_SMBus_host_controller_enabled(), hst_en)

    @patch("chipsec.hal.smbus.iobar")
    def test_enable_SMBus_host_controller(self, mock_iobar):
        base_address = 123456
        mock_cs = MagicMock()
        mock_cs.register.read.return_value = 0
        smbus_hal = SMBus(mock_cs)
        smbus_hal.iobar.is_IO_BAR_defined.return_value = True
        smbus_hal.iobar.get_IO_BAR_base_address.return_value = (base_address, None)
        smbus_hal.enable_SMBus_host_controller()
        self.assertEqual(smbus_hal.cs.register.write.call_count, 2)
        self.assertEqual(smbus_hal.cs.register.read.call_count, 2)

    def test_reset_SMBus_controller_valid(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.read.side_effect = [321, 0x8, 0]
        self.assertTrue(smbus_hal.reset_SMBus_controller())
        self.assertEqual(smbus_hal.cs.register.read.call_count, 3)

    def test_reset_SMBus_controller_invalid(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.read.return_value = 0x8
        self.assertFalse(smbus_hal.reset_SMBus_controller())
        self.assertEqual(smbus_hal.cs.register.read.call_count, SMBUS_POLL_COUNT + 1)

    def test__is_smbus_ready_valid(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.read_field.return_value = 0
        self.assertTrue(smbus_hal._is_smbus_ready())
        self.assertEqual(smbus_hal.cs.register.read_field.call_count, 1)

    def test__is_smbus_ready_invalid(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.read_field.return_value = 1
        self.assertFalse(smbus_hal._is_smbus_ready())
        self.assertEqual(smbus_hal.cs.register.read_field.call_count, SMBUS_POLL_COUNT)

    def test__wait_for_cycle_pass(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.get_field.side_effect = [0, 1, 0]
        self.assertTrue(smbus_hal._wait_for_cycle())

    def test__wait_for_cycle_fail(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        smbus_hal.logger = MagicMock()
        smbus_hal.logger.HAL = True
        mock_cs.register.get_field.side_effect = [1, 1, 1]
        self.assertFalse(smbus_hal._wait_for_cycle())

    def test__wait_for_cycle_invalid_dev_err(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        smbus_hal.logger = MagicMock()
        smbus_hal.logger.HAL = True
        mock_cs.register.has_field.return_value = True
        mock_cs.register.get_field.side_effect = [1, 1, 0, 1]
        self.assertFalse(smbus_hal._wait_for_cycle())

    def test__wait_for_cycle_invalid_bus_err(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        smbus_hal.logger = MagicMock()
        smbus_hal.logger.HAL = True
        mock_cs.register.has_field.side_effect = [False, True]
        mock_cs.register.get_field.side_effect = [1, 1, 0, 1]
        self.assertFalse(smbus_hal._wait_for_cycle())

    def test_read_byte_pass(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        read_field_val = 1234
        mock_cs.register.get_field.side_effect = [0, 1, 0]
        mock_cs.register.read_field.return_value = read_field_val
        self.assertEqual(smbus_hal.read_byte(1, 2), read_field_val)

    def test_read_byte_fail(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.get_field.side_effect = [1, 1, 1]
        self.assertEqual(smbus_hal.read_byte(1, 2), 0xFF)

    def test_write_byte_pass(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.get_field.side_effect = [0, 1, 0]
        self.assertTrue(smbus_hal.write_byte(1, 2, 3))

    def test_write_byte_fail(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.get_field.side_effect = [1, 1, 1]
        self.assertFalse(smbus_hal.write_byte(1, 2, 3))

    def test_read_range(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        size = 1
        mock_cs.register.get_field.side_effect = [0, 1, 0] * size
        mock_cs.register.read_field.return_value = 1
        buffer = bytes([1] * size)
        self.assertEqual(smbus_hal.read_range(1, 1, size), buffer)

    def test_write_range(self):
        mock_cs = MagicMock()
        smbus_hal = SMBus(mock_cs)
        mock_cs.register.get_field.side_effect = [0, 1, 0]
        mock_cs.register.read_field.return_value = 1
        buffer = bytes([1])
        self.assertTrue(smbus_hal.write_range(1, 2, buffer))



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import unittest

from tests.software import mock_helper, util


class TestDescChipsecUtil(util.TestChipsecUtil):
    """Test the Desc commands (gdt, idt, ldt) exposed by chipsec_utils."""

    def test_gdt(self):

        class GDTHelper(mock_helper.TestHelper):

            def get_descriptor_table(self, cpu_thread_id, desc_table_code):
                return (63, 0x1000, 0x0)

            def read_phys_mem(self, pa, length):
                return b"\xff" * length

        self._chipsec_util("gdt 0", GDTHelper)
        self.assertIn(b"# of entries    : 4", self.log)


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
Command-line utility providing access to Intel TXT (Trusted Execution Technology) registers

Usage:
    >>> chipsec_util txt dump
    >>> chipsec_util txt state
"""

from argparse import ArgumentParser
from chipsec.command import BaseCommand, toLoad
from chipsec.library.exceptions import HWAccessViolationError
from chipsec.testcase import ExitCode
import struct


class TXTCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)
        subparsers = parser.add_subparsers()
        parser_state = subparsers.add_parser('dump')
        parser_state.set_defaults(func=self.txt_dump)
        parser_state = subparsers.add_parser('state')
        parser_state.set_defaults(func=self.txt_state)
        parser.parse_args(self.argv, namespace=self)

    def txt_dump(self):
        # Read TXT Public area as hexdump, with absolute address and skipping zeros
        txt_public = self.cs.mem.read_physical_mem(0xfed30000, 0x1000)
        has_skipped_line = False
        for offset in range(0, len(txt_public), 16):
            line_bytes = txt_public[offset:offset + 16]
            if all(b == 0 for b in line_bytes):
                has_skipped_line = True
                continue
            if has_skipped_line:
                self.logger.log("[CHIPSEC] *")
                has_skipped_line = False
            line_hex = " ".join("{:02X}".format(b) for b in line_bytes)
            self.logger.log("[CHIPSEC] {:08X}: {}".format(0xfed30000 + offset, line_hex))

    def _log_register(self, reg_name):
        """Log the content of a register with lines starting with [CHIPSEC]"""
        if not self.cs.register.is_defined(reg_name):
            return
        reg_def = self.cs.register.get_def(reg_name)
        value = self.cs.register.read(reg_name)
        desc = reg_def["desc"]
        if reg_def["type"] == "memory":
            addr = reg_def["address"] + reg_def["offset"]
            desc += ", at {:08X}".format(addr)
        self.logger.log("[CHIPSEC] {} = 0x{:0{width}X} ({})".format(
            reg_name, value, desc, width=reg_def['size'] * 2))

        if 'FIELDS' in reg_def:
            sorted_fields = sorted(reg_def['FIELDS'].items(), key=lambda field: int(field[1]['bit']))
            for field_name, field_attrs in sorted_fields:
                field_bit = int(field_attrs['bit'])
                field_size = int(field_attrs['size'])
                field_mask = (1 << field_size) - 1
                field_value = (value >> field_bit) & field_mask
                self.logger.log("[CHIPSEC]     [{:02d}] {:23} = {:X} << {}".format(
                    field_bit, field_name, field_value, field_attrs['desc']))

    def txt_state(self):
        """Dump Intel TXT state

        This is similar to command "txt-stat" from Trusted Boot project
        https://sourceforge.net/p/tboot/code/ci/v2.0.0/tree/utils/txt-stat.c
        which was documented on
        https://www.intel.com/content/dam/www/public/us/en/documents/guides/dell-one-stop-txt-activation-guide.pdf
        and it is also similar to command "sl-stat" from TrenchBoot project
        https://github.com/TrenchBoot/sltools/blob/842cfd041b7454727b363b72b6d4dcca9c00daca/sl-stat/sl-stat.c
        """
        # Read bits in CPUID
        (eax, ebx, ecx, edx) = self.cs.cpu.cpuid(0x01, 0x00)
        self.logger.log("[CHIPSEC] CPUID.01H.ECX[Bit 6] = {} << Safer Mode Extensions (SMX)".format((ecx >> 6) & 1))
        self.logger.log("[CHIPSEC] CPUID.01H.ECX[Bit 5] = {} << Virtual Machine Extensions (VMX)".format((ecx >> 5) & 1))

        # Read bits in CR4
        cr4 = self.cs.cpu.read_cr(0, 4)
        self.logger.log("[CHIPSEC] CR4.SMXE[Bit 14] = {} << Safer Mode Extensions Enable".format((cr4 >> 14) & 1))
        self.logger.log("[CHIPSEC] CR4.VMXE[Bit 13] = {} << Virtual Machine Extensions Enable".format((cr4 >> 13) & 1))

        # Read bits in MSR IA32_FEATURE_CONTROL
        self._log_register("IA32_FEATURE_CONTROL")
        self.logger.log("[CHIPSEC]")

        # Read TXT Device ID
        self._log_register("TXT_DIDVID")
        self.logger.log("[CHIPSEC]")

        # Read hashes of public keys
        txt_pubkey = struct.pack("<QQQQ",
                                 self.cs.register.read("TXT_PUBLIC_KEY_0"),
                                 self.cs.register.read("TXT_PUBLIC_KEY_1"),
                                 self.cs.register.read("TXT_PUBLIC_KEY_2"),
                                 self.cs.register.read("TXT_PUBLIC_KEY_3"),
                                 )
        self.logger.log("[CHIPSEC] TXT Public Key Hash: {}".format(txt_pubkey.hex()))

        try:
            eax, edx = self.cs.msr.read_msr(0, 0x20)
            pubkey_in_msr = struct.pack("<II", eax, edx)
            eax, edx = self.cs.msr.read_msr(0, 0x21)
            pubkey_in_msr += struct.pack("<II", eax, edx)
            eax, edx = self.cs.msr.read_msr(0, 0x22)
            pubkey_in_msr += struct.pack("<II", eax, edx)
            eax, edx = self.cs.msr.read_msr(0, 0x23)
            pubkey_in_msr += struct.pack("<II", eax, edx)
            self.logger.log("[CHIPSEC] Public Key Hash in MSR[0x20...0x23]: {}".format(pubkey_in_msr.hex()))
        except HWAccessViolationError as exc:
            # Report the exception and continue
            self.logger.log("[CHIPSEC] Unable to read Public Key Hash in MSR[0x20...0x23]: {}".format(exc))
        self.logger.log("[CHIPSEC]")

        # Read TXT status
        self._log_register("TXT_STS")
        self._log_register("TXT_ESTS")
        self._log_register("TXT_E2STS")
        self._log_register("TXT_ERRORCODE")
        self.logger.log("[CHIPSEC]")
        self._log_register("TXT_SPAD")
        self._log_register("TXT_ACM_STATUS")
        self._log_register("TXT_FIT")
        self._log_register("TXT_SCRATCHPAD")
        self.logger.log("[CHIPSEC]")

        # Read memory area for TXT components
        self._log_register("TXT_SINIT_BASE")
        self._log_register("TXT_SINIT_SIZE")
        self._log_register("TXT_MLE_JOIN")
        self._log_register("TXT_HEAP_BASE")
        self._log_register("TXT_HEAP_SIZE")
        self._log_register("TXT_MSEG_BASE")
        self._log_register("TXT_MSEG_SIZE")
        self.logger.log("[CHIPSEC]")

        # Read other registers in the TXT memory area
        self._log_register("TXT_DPR")
        self._log_register("TXT_VER_FSBIF")
        self._log_register("TXT_VER_QPIIF")
        self._log_register("TXT_PCH_DIDVID")
        self._log_register("INSMM")

    def run(self):
        try:
            self.func()
        except Exception:
            self.ExitCode = ExitCode.ERROR


commands = {'txt': TXTCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2018-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
JEDED ID : Manufacturers and Device IDs
"""

from typing import Dict


class JEDEC_ID:

    MANUFACTURER: Dict[int, str] = {0xEF: 'Winbond',
                                    0xC2: 'Macronix'}

    DEVICE: Dict[int, str] = {0xEF4018: 'W25Q128 (SPI)',
                              0xEF6018: 'W25Q128 (QPI)',
                              0xEF4017: 'W25Q64FV (SPI)',
                              0xEF6017: 'W25Q64FV (QPI)',
                              0xEF7016: 'W25Q32JV',
                              0xEF4019: 'W25Q256',
                              0xC22017: 'MX25L6408',
                              0xC22018: 'MX25L12805'}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2017, Google Inc
# Copyright (c) 2019-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

"""
Trusted Platform Module Event Log

Based on the following specifications:

`TCG EFI Platform Specification For TPM Family 1.1 or 1.2 <https://trustedcomputinggroup.org/wp-content/uploads/TCG_EFI_Platform_1_22_Final_-v15.pdf>`_

`TCG PC Client Specific Implementation Specification for Conventional BIOS", version 1.21 <https://trustedcomputinggroup.org/wp-content/uploads/TCG_PCClientImplementation_1-21_1_00.pdf>`_

`TCG EFI Protocol Specification, Family "2.0" <https://trustedcomputinggroup.org/wp-content/uploads/EFI-Protocol-Specification-rev13-160330final.pdf>`_

`TCG PC Client Platform Firmware Profile Specification <https://trustedcomputinggroup.org/wp-content/uploads/PC-ClientSpecific_Platform_Profile_for_TPM_2p0_Systems_v51.pdf>`_
"""

import struct

from typing import Any, Dict, BinaryIO, Optional, Type, TypeVar
from chipsec.library.logger import logger

EventType = TypeVar('EventType', bound='TcgPcrEvent')


class TcgPcrEvent:
    """An Event (TPM 1.2 format) as recorded in the SML."""

    _header_fmt = "II20sI"
    _header_size = struct.calcsize(_header_fmt)

    def __init__(self, pcr_index: int, event_type: int, digest: bytes, event_size: int, event: Any):
        self.pcr_index = pcr_index
        self.event_type = event_type
        name = SML_EVENT_TYPE.get(self.event_type)
        if isinstance(name, str):
            self.event_type_name = name
        self.digest = digest
        self.event_size = event_size
        self.event = event

    @classmethod
    def parse(cls: Type[EventType], log: BinaryIO) -> Optional[EventType]:
        """Try to read an event from the log.

        Args:
            log (file-like): Log where the event is stored.

        Returns:
            An instance of the created event. If a subclass
            exists for such event_type, an object of this class
            is returned. Otherwise, a TcgPcrEvent is returned.
        """
        header = log.read(cls._header_size)
        if not header:
            return None
        fields = struct.unpack(cls._header_fmt, header)
        pcr_index, event_type, digest, event_size = fields
        event = log.read(event_size)
        if len(event) != event_size:
            logger().log_warning("[tpm_eventlog] event data length does not match the expected size")
        name = SML_EVENT_TYPE.get(event_type)
        kls = cls if isinstance(name, str) else name
        if kls is None:
            return None
        return kls(pcr_index, event_type, digest, event_size, event)

    def __str__(self) -> str:
        if self.event_type_name:
            t = self.event_type_name
        else:
            t = f'(0x{self.event_type:x}'
        return f'PCR: {self.pcr_index:d}\ttype: {t.ljust(EVENT_TYPE_MAX_LENGTH)}\tsize: 0x{self.event_size:x}\tdigest: {self.digest.hex()}'


class SCRTMVersion(TcgPcrEvent):
    def __init__(self, *args: Any):
        super(SCRTMVersion, self).__init__(*args)
        self.event_type_name = "EV_S_CRTM_VERSION"
        self.version: bytes = self.event

    def __str__(self) -> str:
        _str = super(SCRTMVersion, self).__str__()
        try:
            _str += f'\n\t+ version: {self.version.decode("utf-16")}'
        except:
            if logger().HAL:
                logger().log_warning("[tpm_eventlog] CRTM Version is not a valid string")
        return _str


class EFIFirmwareBlob(TcgPcrEvent):
    # Although [4] 9.2.5 mentions UNIT64 for the length, [1] 7.7 uses
    # a UINTN. Use a native unsigned long to cover the most general case.
    _event_fmt = "@QL"

    def __init__(self, *args: Any):
        super(EFIFirmwareBlob, self).__init__(*args)
        self.event_type_name = "EV_EFI_PLATFORM_FIRMWARE_BLOB"
        base, length = struct.unpack(self._event_fmt, self.event)
        self.base = base
        self.length = length

    def __str__(self) -> str:
        _blob = super(EFIFirmwareBlob, self).__str__()
        _str = f'{_blob}\n\t+ base: 0x{self.base:x}\tlength: 0x{self.length:x}'
        return _str


SML_EVENT_TYPE: Dict[int, Any] = {
    # From reference [2]
    0x0: "EV_PREBOOT_CERT",
    0x1: "EV_POST_CODE",
    0x2: "EV_UNUSED",
    0x3: "EV_NO_ACTION",
    0x4: "EV_SEPARATOR",
    0x5: "EV_ACTION",
    0x6: "EV_EVENT_TAG",
    0x7: "EV_S_CRTM_CONTENTS",
    0x8: SCRTMVersion,
    0x9: "EV_CPU_MICROCODE",
    0xA: "EV_PLATFORM_CONFIG_FLAGS",
    0xB: "EV_TABLE_OF_DEVICES",
    0xC: "EV_COMPACT_HASH",
    0xD: "EV_IPL",
    0xE: "EV_IPL_PARTITION_DATA",
    0xF: "EV_NONHOST_CODE",
    0x10: "EV_NONHOST_CONFIG",
    0x11: "EV_NONHOST_INFO",
    0x12: "EV_OMIT_BOOT_DEVICE_EVENTS",

    # From reference [1]
    0x80000000: "EV_EFI_EVENT_BASE",
    0x80000001: "EV_EFI_VARIABLE_DRIVER_CONFIG",
    0x80000002: "EV_EFI_VARIABLE_BOOT",
    0x80000003: "EV_EFI_BOOT_SERVICES_APPLICATION",
    0x80000004: "EV_EFI_BOOT_SERVICES_DRIVER",
    0x80000005: "EV_EFI_RUNTIME_SERVICES_DRIVER",
    0x80000006: "EV_EFI_GPT_EVENT",
    0x80000007: "EV_EFI_ACTION",
    0x80000008: EFIFirmwareBlob,
    0x80000009: "EV_EFI_HANDOFF_TABLES",
    0x800000E0: "EV_EFI_VARIABLE_AUTHORITY"
}

EVENT_TYPE_MAX_LENGTH: int = max([len(v) for v in SML_EVENT_TYPE.values()
                             if isinstance(v, str)])


class PcrLogParser:
    """Iterator over the events of a log."""

    def __init__(self, log: BinaryIO):
        self.log = log

    def __iter__(self) -> 'PcrLogParser':
        return self

    def __next__(self) -> TcgPcrEvent:
        event = TcgPcrEvent.parse(self.log)
        if not event:
            raise StopIteration()
        return event

    def next(self) -> TcgPcrEvent:
        return self.__next__()


def parse(log: BinaryIO) -> None:
    """Simple wrapper around PcrLogParser."""
    for event in PcrLogParser(log):
        logger().log(str(event))



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#


"""
The vmem command provides direct access to read and write virtual memory.

>>> chipsec_util vmem <op> <virtual_address> <length> [value|buffer_file]
>>>
>>> <virtual_address> : 64-bit virtual address
>>> <op>               : read|readval|write|writeval|allocate|pagedump|search|getphys
>>> <length>           : byte|word|dword or length of the buffer from <buffer_file>
>>> <value>            : byte, word or dword value to be written to memory at <virtual_address>
>>> <buffer_file>      : file with the contents to be written to memory at <virtual_address>

Examples:

>>> chipsec_util vmem <op>     <virtual_address>  <length> [value|file]
>>> chipsec_util vmem readval  0xFED40000         dword
>>> chipsec_util vmem read     0x41E              0x20     buffer.bin
>>> chipsec_util vmem writeval 0xA0000            dword    0x9090CCCC
>>> chipsec_util vmem write    0x100000000        0x1000   buffer.bin
>>> chipsec_util vmem write    0x100000000        0x10     000102030405060708090A0B0C0D0E0F
>>> chipsec_util vmem allocate                    0x1000
>>> chipsec_util vmem search   0xF0000            0x10000  _SM_
>>> chipsec_util vmem getphys  0xFED00000
"""

import os
import chipsec_util

from chipsec.command import BaseCommand, toLoad
from chipsec.hal import virtmem
from chipsec.library.defines import bytestostring
from chipsec.library.logger import print_buffer_bytes
from chipsec.library.file import write_file, read_file
from argparse import ArgumentParser


# Virtual Memory
class VMemCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_read.add_argument('size', type=lambda x: int(x, 16), nargs='?', default=0x100, help='Length (hex)')
        parser_read.add_argument('buf_file', type=str, nargs='?', default='', help='Buffer file name')
        parser_read.set_defaults(func=self.vmem_read)

        parser_readval = subparsers.add_parser('readval')
        parser_readval.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_readval.add_argument('length', type=str, nargs='?', default=None, help='Length [byte, word, dword] or (hex)')
        parser_readval.set_defaults(func=self.vmem_readval)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_write.add_argument('size', type=lambda x: int(x, 16), default=0x100, help='Length (hex)')
        parser_write.add_argument('buf_file', type=str, help='Buffer file name')
        parser_write.set_defaults(func=self.vmem_write)

        parser_writeval = subparsers.add_parser('writeval')
        parser_writeval.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_writeval.add_argument('length', type=str, help='Length [byte, word, dword] or (hex)')
        parser_writeval.add_argument('value', type=lambda x: int(x, 16), help='Value (hex)')
        parser_writeval.set_defaults(func=self.vmem_writeval)

        parser_search = subparsers.add_parser('search')
        parser_search.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_search.add_argument('size', type=lambda x: int(x, 16), help='Size of memory to search (hex)')
        parser_search.add_argument('value', type=str, help='Value (string)')
        parser_search.set_defaults(func=self.vmem_search)

        parser_allocate = subparsers.add_parser('allocate')
        parser_allocate.add_argument('size', type=lambda x: int(x, 16), help='Size of memory to allocate (hex)')
        parser_allocate.set_defaults(func=self.vmem_allocate)

        parser_getphys = subparsers.add_parser('getphys')
        parser_getphys.add_argument('virt_address', type=lambda x: int(x, 16), help='Address (hex)')
        parser_getphys.set_defaults(func=self.vmem_getphys)
        parser.parse_args(self.argv, namespace=self)

    def set_up(self) -> None:
        self._vmem = virtmem.VirtMemory(self.cs)

    def vmem_read(self):
        self.logger.log('[CHIPSEC] Reading buffer from memory: VA = 0x{:016X}, len = 0x{:X}.'.format(self.virt_address, self.size))
        try:
            buffer = self._vmem.read_virtual_mem(self.virt_address, self.size)
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')
            return

        if self.buf_file:
            write_file(self.buf_file, buffer)
            self.logger.log("[CHIPSEC] Written 0x{:X} bytes to '{}'".format(len(buffer), self.buf_file))
        else:
            print_buffer_bytes(buffer)

    def vmem_readval(self):
        width = 0x4
        value = 0x0
        if self.length is not None:
            if chipsec_util.is_option_valid_width(self.length):
                width = chipsec_util.get_option_width(self.length)
            else:
                try:
                    width = int(self.length, 16)
                except:
                    width = 0

        self.logger.log('[CHIPSEC] Reading {:X}-byte value from VA 0x{:016X}.'.format(width, self.virt_address))
        try:
            if 0x1 == width:
                value = self._vmem.read_virtual_mem_byte(self.virt_address)
            elif 0x2 == width:
                value = self._vmem.read_virtual_mem_word(self.virt_address)
            elif 0x4 == width:
                value = self._vmem.read_virtual_mem_dword(self.virt_address)
            else:
                self.logger.log_error("Must specify <length> argument in 'mem readval' as one of {}".format(chipsec_util.CMD_OPTS_WIDTH))
                return
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')
            return
        self.logger.log('[CHIPSEC] value = 0x{:X}'.format(value))

    def vmem_write(self):
        if not os.path.exists(self.buf_file):
            try:
                buffer = bytearray.fromhex(self.buf_file)
            except ValueError as e:
                self.logger.log_error("Incorrect <value> specified: '{}'".format(self.buf_file))
                self.logger.log_error(str(e))
                return
            self.logger.log("[CHIPSEC] Read 0x{:X} hex bytes from command-line: {}'".format(len(buffer), self.buf_file))
        else:
            buffer = read_file(self.buf_file)
            self.logger.log("[CHIPSEC] Read 0x{:X} bytes from file '{}'".format(len(buffer), self.buf_file))

        if len(buffer) < self.size:
            self.logger.log_error("Number of bytes read (0x{:X}) is less than the specified <length> (0x{:X})".format(len(buffer), self.size))
            return

        self.logger.log('[CHIPSEC] Writing buffer to memory: VA = 0x{:016X}, len = 0x{:X}.'.format(self.virt_address, self.size))
        self._vmem.write_virtual_mem(self.virt_address, self.size, buffer)

    def vmem_writeval(self):
        if chipsec_util.is_option_valid_width(self.length):
            width = chipsec_util.get_option_width(self.length)
        else:
            try:
                width = int(self.length, 16)
            except ValueError:
                width = 0

        self.logger.log('[CHIPSEC] Writing {:X}-byte value 0x{:X} to VA 0x{:016X}..'.format(width, self.value, self.virt_address))
        try:
            if 0x1 == width:
                self._vmem.write_virtual_mem_byte(self.virt_address, self.value)
            elif 0x2 == width:
                self._vmem.write_virtual_mem_word(self.virt_address, self.value)
            elif 0x4 == width:
                self._vmem.write_virtual_mem_dword(self.virt_address, self.value)
            else:
                self.logger.log_error("Must specify <length> argument in 'mem writeval' as one of {}".format(chipsec_util.CMD_OPTS_WIDTH))
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')

    def vmem_search(self):
        try:
            buffer = self._vmem.read_virtual_mem(self.virt_address, self.size)
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')
            return

        buffer = bytestostring(buffer)
        offset = buffer.find(self.value)

        self.logger.log("[CHIPSEC] Search buffer for '{}':".format(self.value))
        self.logger.log('          VA = 0x{:016X}, len = 0x{:X}'.format(self.virt_address, self.size))
        if offset != -1:
            self.logger.log('[CHIPSEC] Target address = 0x{:X}.'.format(self.virt_address + offset))
        else:
            self.logger.log('[CHIPSEC] Could not find the target in the searched range.')

    def vmem_allocate(self):
        try:
            (va, pa) = self._vmem.alloc_virtual_mem(self.size)
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')
            return
        self.logger.log('[CHIPSEC] Allocated {:X} bytes of virtual memory:'.format(self.size))
        self.logger.log('          VA = 0x{:016X}'.format(va))
        self.logger.log('          PA = 0x{:016X}'.format(pa))

    def vmem_getphys(self):
        try:
            pa = self._vmem.va2pa(self.virt_address)
        except (TypeError, OSError):
            self.logger.log_error('Error mapping VA to PA.')
            return
        if pa is not None:
            self.logger.log('[CHIPSEC] Virtual memory:')
            self.logger.log('          VA = 0x{:016X}'.format(self.virt_address))
            self.logger.log('          PA = 0x{:016X}'.format(pa))


commands = {'vmem': VMemCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

# To execute: python[3] -m unittest tests.helpers.test_windowshelper

import unittest
from unittest.mock import patch, Mock
import sys

from chipsec.library.exceptions import UnimplementedAPIError
from tests.helpers.helper_utils import packer

DEBUG = False  # Set to True to print the args passed to the driver
DRIVER_HANDLE = '12345'


class pcibdf_sideeffect():
    def __init__(self, b, d, f, o) -> None:
        self.BUS = b
        self.DEV = d
        self.FUNC = f
        self.OFF = o


def print_args(args):
    for i in args:
        if type(i) is int:
            print(hex(i))
        else:
            print(i)


@patch('chipsec.helper.windows.windowshelper.win32file.CreateFile')
@patch('chipsec.helper.windows.windowshelper.win32file.DeviceIoControl')
class WindowsHelperTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set up mock modules
        cls._mocked_modules = ['pywintypes', 'win32service', 'windll',
                               'winerror', 'win32file', 'win32api', 'win32process',
                               'win32security', 'win32serviceutil', 'ctypes', 'win32.lib']
        for mod in cls._mocked_modules:
            sys.modules[mod] = Mock()

    @classmethod
    def tearDownClass(cls):
        # remove mocked modules
        for mod in cls._mocked_modules:
            del sys.modules[mod]

    ipacker = packer('I')
    qpacker = packer()
    ioctl_dict = {
        (0x22e028, b'\x00\x00\x00\x004\x12\x00\x00\x08\x00\x00\x00abc'):
            b'\x01',
        (0x22e024, b'\x00\x00\x00\x00\x00P\x00\x00\x02\x00\x00\x00'):
            b'\xac\xdc',
        (0x22e054, b'E#\x01\x00\x00\x00\x00\x00'):
            b'\x00\x00\x00\x00\x00\x00\x00\x00',
        (0x22e01c, b'\x00\x00\x00\x00\x00\x00\x00\x00\x01'):
            b'\x86',
        (0x22e01c, b'\x00\x00\x00\x00\x00\x00\x00\x00\x02'):
            b'\x86\x80',
        (0x22e01c, b'\x00\x00\x00\x00\x00\x00\x00\x00\x04'):
            ipacker.custom_pack(1, 0x9a128086, 0),
        (0x22e038, b'\xb8\x00\x01'):
            b'\x11',
        (0x22e038, b'\xb8\x00\x02'):
            b'\x22\x11',
        (0x22e038, b'\xb8\x00\x04'):
            b'\x44\x33\x22\x11',
        (0x22e04c, b'\x01\x00\x00\x00\x00\x00\x00\x00'):
            ipacker.pack_cpuinfo(0x406F1),
        (0x22e070, b'\x00\x00\x00\x00\x00\x00\x01\xfe\x04\x00\x00\x00'):
            ipacker.custom_pack(1, 0x7FF0200, 0),
        (0x22e048, b'\x00\x00\x00\x00\x00\x10\x00\x00\x08\x00\x00\x00'):
            qpacker.custom_pack(2, 1, 0),
        (0x22e064, b'\x00\x00\x00\x00\x00\x00'):
            b'3\x00\x05\x80\x00\x00\x00\x00',
        (0x22e060, b'\x00\x003\x00\x05\x80\x00\x00\x00\x00\x00\x00\x00\x00'):
            b'',
        (0x22e034, b'\x00\x00\x00\x00:\x00\x00\x00'):
            b'\x01\x00\x00\x00\x00\x00\x00\x00',
        (0x22e030, b'\x00\x00\x00\x00:\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00'):
            b'',
        (0x22e040, b'\x00\x00\x00\x00\x01'):
            b'W\x00\xb0\x1f\xeav\x04\xf8\xff\xff\xb0\xbf0\x03\x00\x00\x00\x00',
    }

    def ioctlret(*args):
        if DEBUG:
            print_args(args)
        if args[1:3] in WindowsHelperTest.ioctl_dict:
            return WindowsHelperTest.ioctl_dict[args[1:3]]
        return b'\xab'*40

    @patch('chipsec.helper.windows.windowshelper.windll.ntdll.NtQuerySystemInformation')
    @patch('chipsec.helper.windows.windowshelper.logger')
    @patch('chipsec.helper.windows.windowshelper.win32file')
    @patch('chipsec.helper.windows.windowshelper.win32serviceutil')
    @patch('chipsec.helper.windows.windowshelper.win32service')
    @patch('chipsec.helper.windows.windowshelper.os.path.isfile')
    @patch('chipsec.helper.windows.windowshelper.platform')
    def setUp(self, mock_platform, os_path_isfile, mock_win32sservice, mock_win32serviceutil, mock_win32file, mock_logger, mock_NtQuerySystemInformation):
        unittest.TestCase.setUp(self)
        mock_platform.system.return_value = "windows"
        import chipsec.helper.windows.windowshelper as wh
        wh.win32service.error = Exception
        wh.win32security.TOKEN_READ = 0x20008
        wh.win32security.TOKEN_ADJUST_PRIVILEGES = 0x20
        wh.win32serviceutil.QueryServiceStatus.return_value = (16, 4, 197, 0, 0, 0, 0)
        wh.win32service.SERVICE_STOPPED = 1
        wh.win32service.SERVICE_RUNNING = 4
        wh.win32service.SERVICE_QUERY_STATUS = 4
        wh.win32service.SERVICE_START = 16
        wh.win32service.SERVICE_STOP = 32
        wh.win32service.SERVICE_KERNEL_DRIVER = 1
        wh.win32service.SERVICE_DEMAND_START = 3
        wh.win32service.SERVICE_ERROR_NORMAL = 1
        wh.FILE_SHARE_READ = 1
        wh.FILE_SHARE_WRITE = 2
        wh.OPEN_EXISTING = 3
        wh.FILE_ATTRIBUTE_NORMAL = 0x80
        wh.FILE_FLAG_OVERLAPPED = 0x40000000
        wh.INVALID_HANDLE_VALUE = -1
        wh.c_char = type('', (object,), {"__mul__": lambda self, other: Mock()})()
        os_path_isfile().return_value = True
        mock_win32sservice.CreateService.return_value = DRIVER_HANDLE
        mock_win32sservice.CloseServiceHandle().return_value = None
        self.wh = wh
        with patch.dict(wh.os.__dict__, {'chown': lambda *args: None}):
            self.whelper = wh.WindowsHelper()
            self.assertTrue(self.whelper.create())
            self.assertTrue(self.whelper.start())

    @patch('chipsec.helper.windows.windowshelper.win32api.CloseHandle')
    def tearDown(self, _):
        unittest.TestCase.tearDown(self)
        self.assertTrue(self.whelper.delete())
        self.whelper.__del__()

    # TODO def test_map_io_space(self, _):

    def _assign_mocks(self, mocks):
        mock_ioctl = mocks[0]
        mock_createfile = mocks[1]
        mock_ioctl.side_effect = WindowsHelperTest.ioctlret
        mock_createfile.return_value = DRIVER_HANDLE

    def test_write_phys_mem(self, *mocks):
        self._assign_mocks(mocks)
        retval = self.whelper.write_phys_mem(0x1234, 0x8, "abc")
        self.assertEqual(retval, 0x1)

    def test_read_phys_mem(self, *mocks):
        self._assign_mocks(mocks)
        mem_value = self.whelper.read_phys_mem(0x5000, 0x2)
        self.assertEqual(mem_value, b'\xac\xdc')

    def test_va2pa(self, *mocks):
        self._assign_mocks(mocks)
        pa = self.whelper.va2pa(0x12345)
        self.assertEqual(pa, (0, 0))

    @patch('chipsec.helper.windows.windowshelper.PCI_BDF')
    def test_read_pci_reg_cpu_one_byte(self, pci_bdf, *mocks):
        self._assign_mocks(mocks)
        pci_bdf.side_effect = pcibdf_sideeffect
        pci_read_value = self.whelper.read_pci_reg(0, 0, 0, 0, 0x1)
        self.assertEqual(pci_read_value, 0x86)

    @patch('chipsec.helper.windows.windowshelper.PCI_BDF')
    def test_read_pci_reg_cpu_two_bytes(self, pci_bdf, *mocks):
        self._assign_mocks(mocks)
        pci_bdf.side_effect = pcibdf_sideeffect
        pci_read_value = self.whelper.read_pci_reg(0, 0, 0, 0, 0x2)
        self.assertEqual(pci_read_value, 0x8086)

    @patch('chipsec.helper.windows.windowshelper.PCI_BDF')
    def test_read_pci_reg_cpu_four_bytes(self, pci_bdf, *mocks):
        self._assign_mocks(mocks)
        pci_bdf.side_effect = pcibdf_sideeffect
        pci_read_value = self.whelper.read_pci_reg(0, 0, 0, 0, 0x4)
        self.assertEqual(pci_read_value, 0x9A128086)

    @patch('chipsec.helper.windows.windowshelper.PCI_BDF')
    def test_write_pci_reg_cpu_one_byte(self, pci_bdf, *mocks):
        self._assign_mocks(mocks)
        pci_bdf.side_effect = pcibdf_sideeffect
        pci_write_return = self.whelper.write_pci_reg(0, 0, 0, 0, 0x1, 0xab)
        self.assertTrue(pci_write_return)

    def test_load_ucode_update(self, *mocks):
        self._assign_mocks(mocks)
        load_ucode_update_return = self.whelper.load_ucode_update(0, b'\x55')
        self.assertTrue(load_ucode_update_return)

    def test_read_io_port_1byte(self, *mocks):
        self._assign_mocks(mocks)
        ioport_read_value = self.whelper.read_io_port(0xB8, 1)
        self.assertEqual(ioport_read_value, 0x11)

    def test_read_io_port_2bytes(self, *mocks):
        self._assign_mocks(mocks)
        ioport_read_value = self.whelper.read_io_port(0xB8, 2)
        self.assertEqual(ioport_read_value, 0x1122)

    def test_read_io_port_4bytes(self, *mocks):
        self._assign_mocks(mocks)
        ioport_read_value = self.whelper.read_io_port(0xB8, 4)
        self.assertEqual(ioport_read_value, 0x11223344)

    def test_write_io_port(self, *mocks):
        self._assign_mocks(mocks)
        ioport_write_return = self.whelper.write_io_port(0xB8, 0x55, 4)
        self.assertTrue(ioport_write_return)

    def test_read_cr(self, *mocks):
        self._assign_mocks(mocks)
        read_cr_return = self.whelper.read_cr(0, 0)
        self.assertEqual(read_cr_return, 0x80050033)

    def test_write_cr(self, *mocks):
        self._assign_mocks(mocks)
        write_cr_return = self.whelper.write_cr(0, 0, 0x80050033)
        self.assertTrue(write_cr_return)

    def test_read_msr(self, *mocks):
        self._assign_mocks(mocks)
        read_msr_return = self.whelper.read_msr(0, 0x3A)
        self.assertEqual(read_msr_return, (1, 0))

    def test_write_msr(self, *mocks):
        self._assign_mocks(mocks)
        write_msr_return = self.whelper.write_msr(0, 0x3a, 1, 0)
        self.assertTrue(write_msr_return)

    def test_get_descriptor_table(self, *mocks):
        self._assign_mocks(mocks)
        get_des_table_return = self.whelper.get_descriptor_table(0, 1)
        self.assertEqual(get_des_table_return,  (0x57, 0xfffff80476ea1fb0, 0x330bfb0))

    def test_cpuid_one(self, *mocks):
        self._assign_mocks(mocks)
        cpuid_value = self.whelper.cpuid(1, 0)
        self.assertEqual(cpuid_value, (0x406F1, 0, 0, 0))

    def test_alloc_phys_mem(self, *mocks):
        self._assign_mocks(mocks)
        alloc_return = self.whelper.alloc_phys_mem(0x8, 0x1000_0000_0000)
        self.assertEqual(alloc_return, (0x1, 0x0))

    def test_free_phys_mem(self, *mocks):
        self._assign_mocks(mocks)
        free_return = self.whelper.free_phys_mem(0x1600_0000)
        self.assertIsNone(free_return)

    def test_read_mmio_reg(self, *mocks):
        self._assign_mocks(mocks)
        mmio_read_value = self.whelper.read_mmio_reg(0xfe010000, 0x4)
        self.assertEqual(mmio_read_value, 0x7FF0200)

    def test_write_mmio_reg(self, *mocks):
        self._assign_mocks(mocks)
        self.whelper.write_mmio_reg(0x123, 0x1, 0x22)

    # TODO def test_get_ACPI_SDT(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_get_ACPI_table(self, *mocks):
        # self._assign_mocks(mocks)

    def test_msgbus_send_read_message(self, *_):
        self.assertRaises(UnimplementedAPIError, self.whelper.msgbus_send_read_message, 0x1, 0x2)

    def test_msgbus_send_write_message(self, *_):
        self.assertRaises(UnimplementedAPIError, self.whelper.msgbus_send_write_message, 0x1, 0x2, 3)

    def test_msgbus_send_message(self, *_):
        self.assertRaises(UnimplementedAPIError, self.whelper.msgbus_send_message, 0x1, 0x2, 3)

    # TODO def test_get_affinity(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_set_affinity(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_EFI_supported(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_delete_EFI_variable(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_list_EFI_variables(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_get_EFI_variable(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_set_EFI_variable(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_hypercall(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_send_sw_smi(self, *mocks):
        # self._assign_mocks(mocks)

    # TODO def test_get_tool_info(self, *mocks):
        # self._assign_mocks(mocks)

    @patch('chipsec.helper.windows.windowshelper.kernel32.GetActiveProcessorCount')
    @patch('chipsec.helper.windows.windowshelper.kernel32.GetActiveProcessorGroupCount')
    def test_get_threads_count(self, mock_kernal32_GetActiveProcessorGroupCount, mock_kernal32_GetActiveProcessorCount, *_):
        mock_kernal32_GetActiveProcessorGroupCount.return_value = 0xffff0004
        mock_kernal32_GetActiveProcessorCount.return_value = 2
        thread_count = self.whelper.get_threads_count()
        self.assertEqual(thread_count, 0x8)

    @patch('chipsec.helper.windows.windowshelper.sizeof')
    @patch('chipsec.helper.windows.windowshelper.addressof')
    @patch('chipsec.helper.windows.windowshelper.c_uint32')
    def test_retpoline_enabled(self, mock_c_uint32, *_):
        mock_c_uint32.return_value = type('', (object,), {"value": 0x4000})()
        self.assertEqual(self.whelper.retpoline_enabled(), True)

# test_reg_get_control
# test_spi_dump
# test_spi_info



import random
import struct


########################################################################################################################


class base_primitive:
    '''
    The primitive base class implements common functionality shared across most primitives.
    '''

    def __init__(self):
        self.fuzz_complete = False     # this flag is raised when the mutations are exhausted.
        self.fuzz_library = []        # library of static fuzz heuristics to cycle through.
        self.fuzzable = True      # flag controlling whether or not the given primitive is to be fuzzed.
        self.mutant_index = 0         # current mutation index into the fuzz library.
        self.original_value = None      # original value of primitive.
        self.rendered = ""        # rendered value of primitive.
        self.value = None      # current value of primitive.

    def exhaust(self):
        '''
        Exhaust the possible mutations for this primitive.

        @rtype:  Integer
        @return: The number of mutations to reach exhaustion
        '''

        num = self.num_mutations() - self.mutant_index

        self.fuzz_complete = True
        self.mutant_index = self.num_mutations()
        self.value = self.original_value

        return num

    def mutate(self):
        '''
        Mutate the primitive by stepping through the fuzz library, return False on completion.

        @rtype:  Boolean
        @return: True on success, False otherwise.
        '''

        # if we've ran out of mutations, raise the completion flag.
        if self.mutant_index == self.num_mutations():
            self.fuzz_complete = True

        # if fuzzing was disabled or complete, and mutate() is called, ensure the original value is restored.
        if not self.fuzzable or self.fuzz_complete:
            self.value = self.original_value
            return False

        # update the current value from the fuzz library.
        self.value = self.fuzz_library[self.mutant_index]

        # increment the mutation count.
        self.mutant_index += 1

        return True

    def num_mutations(self):
        '''
        Calculate and return the total number of mutations for this individual primitive.

        @rtype:  Integer
        @return: Number of mutated forms this primitive can take
        '''

        return len(self.fuzz_library)

    def render(self):
        '''
        Nothing fancy on render, simply return the value.
        '''

        self.rendered = self.value
        return self.rendered

    def reset(self):
        '''
        Reset this primitive to the starting mutation state.
        '''

        self.fuzz_complete = False
        self.mutant_index = 0
        self.value = self.original_value


########################################################################################################################
class delim(base_primitive):
    def __init__(self, value, fuzzable=True, name=None):
        '''
        Represent a delimiter such as :,\r,\n, ,=,>,< etc... Mutations include repetition, substitution and exclusion.

        @type  value:    Character
        @param value:    Original value
        @type  fuzzable: Boolean
        @param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
        @type  name:     String
        @param name:     (Optional, def=None) Specifying a name gives you direct access to a primitive
        '''
        super(delim, self).__init__()
        self.value = self.original_value = value
        self.fuzzable = fuzzable
        self.name = name

        self.s_type = "delim"   # for ease of object identification

        #
        # build the library of fuzz heuristics.
        #

        # if the default delim is not blank, repeat it a bunch of times.
        if self.value:
            self.fuzz_library.append(self.value * 2)
            self.fuzz_library.append(self.value * 5)
            self.fuzz_library.append(self.value * 10)
            self.fuzz_library.append(self.value * 25)
            self.fuzz_library.append(self.value * 100)
            self.fuzz_library.append(self.value * 500)
            self.fuzz_library.append(self.value * 1000)

        # try omitting the delimiter.
        self.fuzz_library.append("")

        # if the delimiter is a space, try throwing out some tabs.
        if self.value == " ":
            self.fuzz_library.append("\t")
            self.fuzz_library.append("\t" * 2)
            self.fuzz_library.append("\t" * 100)

        # toss in some other common delimiters:
        self.fuzz_library.append(" ")
        self.fuzz_library.append("\t")
        self.fuzz_library.append("\t " * 100)
        self.fuzz_library.append("\t\r\n" * 100)
        self.fuzz_library.append("!")
        self.fuzz_library.append("@")
        self.fuzz_library.append("#")
        self.fuzz_library.append("$")
        self.fuzz_library.append("%")
        self.fuzz_library.append("^")
        self.fuzz_library.append("&")
        self.fuzz_library.append("*")
        self.fuzz_library.append("(")
        self.fuzz_library.append(")")
        self.fuzz_library.append("-")
        self.fuzz_library.append("_")
        self.fuzz_library.append("+")
        self.fuzz_library.append("=")
        self.fuzz_library.append(":")
        self.fuzz_library.append(": " * 100)
        self.fuzz_library.append(":7" * 100)
        self.fuzz_library.append(";")
        self.fuzz_library.append("'")
        self.fuzz_library.append("\"")
        self.fuzz_library.append("/")
        self.fuzz_library.append("\\")
        self.fuzz_library.append("?")
        self.fuzz_library.append("<")
        self.fuzz_library.append(">")
        self.fuzz_library.append(".")
        self.fuzz_library.append(",")
        self.fuzz_library.append("\r")
        self.fuzz_library.append("\n")
        self.fuzz_library.append("\r\n" * 64)
        self.fuzz_library.append("\r\n" * 128)
        self.fuzz_library.append("\r\n" * 512)


########################################################################################################################
class group(base_primitive):
    def __init__(self, name, values):
        '''
        This primitive represents a list of static values, stepping through each one on mutation. You can tie a block
        to a group primitive to specify that the block should cycle through all possible mutations for *each* value
        within the group. The group primitive is useful for example for representing a list of valid opcodes.

        @type  name:   String
        @param name:   Name of group
        @type  values: List or raw data
        @param values: List of possible raw values this group can take.
        '''
        super(group, self).__init__()
        self.name = name
        self.values = values
        self.fuzzable = True

        self.s_type = "group"
        self.value = self.values[0]
        self.original_value = self.values[0]

        # sanity check that values list only contains strings (or raw data)
        if self.values != []:
            for val in self.values:
                assert isinstance(val, str), "Value list may only contain strings or raw data"

    def mutate(self):
        '''
        Move to the next item in the values list.

        @rtype:  False
        @return: False
        '''

        if self.mutant_index == self.num_mutations():
            self.fuzz_complete = True

        # if fuzzing was disabled or complete, and mutate() is called, ensure the original value is restored.
        if not self.fuzzable or self.fuzz_complete:
            self.value = self.values[0]
            return False

        # step through the value list.
        self.value = self.values[self.mutant_index]

        # increment the mutation count.
        self.mutant_index += 1

        return True

    def num_mutations(self):
        '''
        Number of values in this primitive.

        @rtype:  Integer
        @return: Number of values in this primitive.
        '''

        return len(self.values)


########################################################################################################################
class random_data(base_primitive):
    def __init__(self, value, min_length, max_length, max_mutations=25, fuzzable=True, step=None, name=None):
        '''
        Generate a random chunk of data while maintaining a copy of the original. A random length range can be specified.
        For a static length, set min/max length to be the same.

        @type  value:         Raw
        @param value:         Original value
        @type  min_length:    Integer
        @param min_length:    Minimum length of random block
        @type  max_length:    Integer
        @param max_length:    Maximum length of random block
        @type  max_mutations: Integer
        @param max_mutations: (Optional, def=25) Number of mutations to make before reverting to default
        @type  fuzzable:      Boolean
        @param fuzzable:      (Optional, def=True) Enable/disable fuzzing of this primitive
        @type  step:          Integer
        @param step:          (Optional, def=None) If not null, step count between min and max reps, otherwise random
        @type  name:          String
        @param name:          (Optional, def=None) Specifying a name gives you direct access to a primitive
        '''

        super(random_data, self).__init__()
        self.value = self.original_value = str(value)
        self.min_length = min_length
        self.max_length = max_length
        self.max_mutations = max_mutations
        self.fuzzable = fuzzable
        self.step = step
        self.name = name

        self.s_type = "random_data"  # for ease of object identification

        if self.step:
            self.max_mutations = (self.max_length - self.min_length) // self.step + 1

    def mutate(self):
        '''
        Mutate the primitive value returning False on completion.

        @rtype:  Boolean
        @return: True on success, False otherwise.
        '''

        # if we've ran out of mutations, raise the completion flag.
        if self.mutant_index == self.num_mutations():
            self.fuzz_complete = True

        # if fuzzing was disabled or complete, and mutate() is called, ensure the original value is restored.
        if not self.fuzzable or self.fuzz_complete:
            self.value = self.original_value
            return False

        # select a random length for this string.
        if not self.step:
            length = random.randint(self.min_length, self.max_length)
        # select a length function of the mutant index and the step.
        else:
            length = self.min_length + self.mutant_index * self.step

        # reset the value and generate a random string of the determined length.
        self.value = b""
        for i in range(length):
            self.value += struct.pack("B", random.randint(0, 255))

        # increment the mutation count.
        self.mutant_index += 1

        return True

    def num_mutations(self):
        '''
        Calculate and return the total number of mutations for this individual primitive.

        @rtype:  Integer
        @return: Number of mutated forms this primitive can take
        '''

        return self.max_mutations


########################################################################################################################
class static(base_primitive):
    def __init__(self, value, name=None):
        '''
        Primitive that contains static content.

        @type  value: Raw
        @param value: Raw static data
        @type  name:  String
        @param name:  (Optional, def=None) Specifying a name gives you direct access to a primitive
        '''

        super(static, self).__init__()
        self.value = self.original_value = value
        self.name = name
        self.fuzzable = False       # every primitive needs this attribute.
        self.s_type = "static"    # for ease of object identification
        self.fuzz_complete = True

    def mutate(self):
        '''
        Do nothing.

        @rtype:  False
        @return: False
        '''

        return False

    def num_mutations(self):
        '''
        Return 0.

        @rtype:  0
        @return: 0
        '''

        return 0


########################################################################################################################
class string(base_primitive):
    # store fuzz_library as a class variable to avoid copying the ~70MB structure across each instantiated primitive.
    fuzz_library = []

    def __init__(self, value, size=-1, padding="\x00", encoding="ascii", fuzzable=True, max_len=0, name=None):
        '''
        Primitive that cycles through a library of "bad" strings. The class variable 'fuzz_library' contains a list of
        smart fuzz values global across all instances. The 'this_library' variable contains fuzz values specific to
        the instantiated primitive. This allows us to avoid copying the near ~70MB fuzz_library data structure across
        each instantiated primitive.

        @type  value:    String
        @param value:    Default string value
        @type  size:     Integer
        @param size:     (Optional, def=-1) Static size of this field, leave -1 for dynamic.
        @type  padding:  Character
        @param padding:  (Optional, def="\\x00") Value to use as padding to fill static field size.
        @type  encoding: String
        @param encoding: (Optonal, def="ascii") String encoding, ex: utf_16_le for Microsoft Unicode.
        @type  fuzzable: Boolean
        @param fuzzable: (Optional, def=True) Enable/disable fuzzing of this primitive
        @type  max_len:  Integer
        @param max_len:  (Optional, def=0) Maximum string length
        @type  name:     String
        @param name:     (Optional, def=None) Specifying a name gives you direct access to a primitive
        '''

        super(string, self).__init__()
        self.value = self.original_value = value
        self.size = size
        self.padding = padding
        self.encoding = encoding
        self.fuzzable = fuzzable
        self.name = name

        self.s_type = "string"  # for ease of object identification

        # add this specific primitives repitition values to the unique fuzz library.
        self.this_library = \
            [
                self.value * 2,
                self.value * 10,
                self.value * 100,

                # UTF-8
                self.value * 2 + "\xfe",
                self.value * 10 + "\xfe",
                self.value * 100 + "\xfe",
            ]

        # if the fuzz library has not yet been initialized, do so with all the global values.
        if not self.fuzz_library:
            string.fuzz_library = \
                [
                    # omission.
                    "",

                    # strings ripped from spike (and some others I added)
                    "/.:/" + "A" * 5000 + "\x00\x00",
                    "/.../" + "A" * 5000 + "\x00\x00",
                    "/.../.../.../.../.../.../.../.../.../.../",
                    "/../../../../../../../../../../../../etc/passwd",
                    "/../../../../../../../../../../../../boot.ini",
                    "..:..:..:..:..:..:..:..:..:..:..:..:..:",
                    "\\\\*",
                    "\\\\?\\",
                    "/\\" * 5000,
                    "/." * 5000,
                    "!@#$%%^#$%#$@#$%$$@#$%^^**(()",
                    "%01%02%03%04%0a%0d%0aADSF",
                    "%01%02%03@%04%0a%0d%0aADSF",
                    "/%00/",
                    "%00/",
                    "%00",
                    "%u0000",
                    "%\xfe\xf0%\x00\xff",
                    "%\xfe\xf0%\x01\xff" * 20,

                    # format strings.
                    "%n" * 100,
                    "%n" * 500,
                    "\"%n\"" * 500,
                    "%s" * 100,
                    "%s" * 500,
                    "\"%s\"" * 500,

                    # command injection.
                    "|touch /tmp/SULLEY",
                    ";touch /tmp/SULLEY;",
                    "|notepad",
                    ";notepad;",
                    "\nnotepad\n",

                    # SQL injection.
                    "1;SELECT%20*",
                    "'sqlattempt1",
                    "(sqlattempt2)",
                    "OR%201=1",

                    # some binary strings.
                    "\xde\xad\xbe\xef",
                    "\xde\xad\xbe\xef" * 10,
                    "\xde\xad\xbe\xef" * 100,
                    "\xde\xad\xbe\xef" * 1000,
                    "\xde\xad\xbe\xef" * 10000,
                    "\x00" * 1000,

                    # miscellaneous.
                    "\r\n" * 100,
                    "<>" * 500,         # sendmail crackaddr (http://lsd-pl.net/other/sendmail.txt)
                ]

            # add some long strings.
            self.add_long_strings("A")
            self.add_long_strings("B")
            self.add_long_strings("1")
            self.add_long_strings("2")
            self.add_long_strings("3")
            self.add_long_strings("<")
            self.add_long_strings(">")
            self.add_long_strings("'")
            self.add_long_strings("\"")
            self.add_long_strings("/")
            self.add_long_strings("\\")
            self.add_long_strings("?")
            self.add_long_strings("=")
            self.add_long_strings("a=")
            self.add_long_strings("&")
            self.add_long_strings(".")
            self.add_long_strings(",")
            self.add_long_strings("(")
            self.add_long_strings(")")
            self.add_long_strings("]")
            self.add_long_strings("[")
            self.add_long_strings("%")
            self.add_long_strings("*")
            self.add_long_strings("-")
            self.add_long_strings("+")
            self.add_long_strings("{")
            self.add_long_strings("}")
            self.add_long_strings("\x14")
            self.add_long_strings("\xFE")   # expands to 4 characters under utf16
            self.add_long_strings("\xFF")   # expands to 4 characters under utf16

            # add some long strings with null bytes thrown in the middle of it.
            for length in [128, 256, 1024, 2048, 4096, 32767, 0xFFFF]:
                s = "B" * length
                s = s[:len(s) // 2] + "\x00" + s[len(s) // 2:]
                string.fuzz_library.append(s)

            # if the optional file '.fuzz_strings' is found, parse each line as a new entry for the fuzz library.
            try:
                fh = open(".fuzz_strings", "r")

                for fuzz_string in fh.readlines():
                    fuzz_string = fuzz_string.rstrip("\r\n")

                    if fuzz_string != "":
                        string.fuzz_library.append(fuzz_string)

                fh.close()
            except:
                pass

        # delete strings which length is greater than max_len.
        if max_len > 0:
            if any(len(s) > max_len for s in self.this_library):
                self.this_library = list(set([s[:max_len] for s in self.this_library]))

            if any(len(s) > max_len for s in self.fuzz_library):
                self.fuzz_library = list(set([s[:max_len] for s in self.fuzz_library]))

    def add_long_strings(self, sequence):
        '''
        Given a sequence, generate a number of selectively chosen strings lengths of the given sequence and add to the
        string heuristic library.

        @type  sequence: String
        @param sequence: Sequence to repeat for creation of fuzz strings.
        '''

        for length in [128, 255, 256, 257, 511, 512, 513, 1023, 1024, 2048, 2049, 4095, 4096, 4097, 5000, 10000, 20000,
                       32762, 32763, 32764, 32765, 32766, 32767, 32768, 32769, 0xFFFF - 2, 0xFFFF - 1, 0xFFFF, 0xFFFF + 1,
                       0xFFFF + 2, 99999, 100000, 500000, 1000000]:

            long_string = sequence * length
            string.fuzz_library.append(long_string)

    def mutate(self):
        '''
        Mutate the primitive by stepping through the fuzz library extended with the "this" library, return False on
        completion.

        @rtype:  Boolean
        @return: True on success, False otherwise.
        '''

        # loop through the fuzz library until a suitable match is found.
        while 1:
            # if we've ran out of mutations, raise the completion flag.
            if self.mutant_index == self.num_mutations():
                self.fuzz_complete = True

            # if fuzzing was disabled or complete, and mutate() is called, ensure the original value is restored.
            if not self.fuzzable or self.fuzz_complete:
                self.value = self.original_value
                return False

            # update the current value from the fuzz library.
            self.value = (self.fuzz_library + self.this_library)[self.mutant_index]

            # increment the mutation count.
            self.mutant_index += 1

            # if the size parameter is disabled, break out of the loop right now.
            if self.size == -1:
                break

            # ignore library items greater then user-supplied length.
            # TODO: might want to make this smarter.
            if len(self.value) > self.size:
                continue

            # pad undersized library items.
            if len(self.value) < self.size:
                self.value = self.value + self.padding * (self.size - len(self.value))
                break

        return True

    def num_mutations(self):
        '''
        Calculate and return the total number of mutations for this individual primitive.

        @rtype:  Integer
        @return: Number of mutated forms this primitive can take
        '''

        return len(self.fuzz_library) + len(self.this_library)

    def render(self):
        '''
        Render the primitive, encode the string according to the specified encoding.
        '''

        # try to encode the string properly and fall back to the default value on failure.
        try:
            self.rendered = str(self.value).encode(self.encoding)
        except:
            self.rendered = str(self.value).encode('latin-1')

        return self.rendered


########################################################################################################################
class bit_field(base_primitive):
    def __init__(self, value, width, max_num=None, endian="<", format="binary", signed=False, full_range=False, fuzzable=True, name=None):
        '''
        The bit field primitive represents a number of variable length and is used to define all other integer types.

        @type  value:      Integer
        @param value:      Default integer value
        @type  width:      Integer
        @param width:      Width of bit fields
        @type  endian:     Character
        @param endian:     (Optional, def=LITTLE_ENDIAN) Endianess of the bit field (LITTLE_ENDIAN: <, BIG_ENDIAN: >)
        @type  format:     String
        @param format:     (Optional, def=binary) Output format, "binary" or "ascii"
        @type  signed:     Boolean
        @param signed:     (Optional, def=False) Make size signed vs. unsigned (applicable only with format="ascii")
        @type  full_range: Boolean
        @param full_range: (Optional, def=False) If enabled the field mutates through *all* possible values.
        @type  fuzzable:   Boolean
        @param fuzzable:   (Optional, def=True) Enable/disable fuzzing of this primitive
        @type  name:       String
        @param name:       (Optional, def=None) Specifying a name gives you direct access to a primitive
        '''

        super(bit_field, self).__init__()
        assert isinstance(width, int)

        if isinstance(value, (int, list, tuple)):
            self.value = self.original_value = value
        else:
            raise ValueError("The supplied value must be either an Int, Long, List or Tuple.")

        self.width = width
        self.max_num = max_num
        self.endian = endian
        self.format = format
        self.signed = signed
        self.full_range = full_range
        self.fuzzable = fuzzable
        self.name = name

        self.rendered = b""        # rendered value
        self.cyclic_index = 0         # when cycling through non-mutating values

        if self.max_num is None:
            self.max_num = self.to_decimal("1" + "0" * width)

        assert isinstance(self.max_num, int)

        # build the fuzz library.
        if self.full_range:
            # add all possible values.
            for i in range(0, self.max_num):
                self.fuzz_library.append(i)
        else:
            if isinstance(value, (list, tuple)):
                # Use the supplied values as the fuzz library.
                for val in value:
                    self.fuzz_library.append(val)
            else:
                # try only "smart" values.
                self.add_integer_boundaries(0)
                self.add_integer_boundaries(self.max_num // 2)
                self.add_integer_boundaries(self.max_num // 3)
                self.add_integer_boundaries(self.max_num // 4)
                self.add_integer_boundaries(self.max_num // 8)
                self.add_integer_boundaries(self.max_num // 16)
                self.add_integer_boundaries(self.max_num // 32)
                self.add_integer_boundaries(self.max_num)

        # if the optional file '.fuzz_ints' is found, parse each line as a new entry for the fuzz library.
        try:
            fh = open(".fuzz_ints", "r")

            for fuzz_int in fh.readlines():
                # convert the line into an integer, continue on failure.
                try:
                    fuzz_int = int(fuzz_int, 16)
                except:
                    continue

                if fuzz_int < self.max_num:
                    self.fuzz_library.append(fuzz_int)

            fh.close()
        except:
            pass

    def add_integer_boundaries(self, integer):
        '''
        Add the supplied integer and border cases to the integer fuzz heuristics library.

        @type  integer: Int
        @param integer: Integer to append to fuzz heuristics
        '''

        for i in range(-10, 10):
            case = integer + i

            # ensure the border case falls within the valid range for this field.
            if 0 <= case < self.max_num:
                if case not in self.fuzz_library:
                    self.fuzz_library.append(case)

    def render(self):
        '''
        Render the primitive.
        '''

        #
        # binary formatting.
        #

        if self.format == "binary":
            bit_stream = ""
            rendered = b""

            # pad the bit stream to the next byte boundary.
            if self.width % 8 == 0:
                bit_stream += self.to_binary()
            else:
                bit_stream = "0" * (8 - (self.width % 8))
                bit_stream += self.to_binary()

            # convert the bit stream from a string of bits into raw bytes.
            for i in range(len(bit_stream) // 8):
                chunk = bit_stream[8 * i:8 * i + 8]
                rendered += struct.pack("B", self.to_decimal(chunk))

            # if necessary, convert the endianess of the raw bytes.
            if self.endian == "<":
                rendered = rendered[::-1]

            self.rendered = rendered

        #
        # ascii formatting.
        #

        else:
            # if the sign flag is raised and we are dealing with a signed integer (first bit is 1).
            if self.signed and self.to_binary()[0] == "1":
                max_num = self.to_decimal("1" + "0" * (self.width - 1))

                # mask off the sign bit.
                val = self.value & self.to_decimal("1" * (self.width - 1))

                # account for the fact that the negative scale works backwards.
                val = max_num - val - 1

                # toss in the negative sign.
                self.rendered = "%d" % ~val

            # unsigned integer or positive signed integer.
            else:
                self.rendered = "%d" % self.value
        return self.rendered

    def to_binary(self, number=None, bit_count=None):
        '''
        Convert a number to a binary string.

        @type  number:    Integer
        @param number:    (Optional, def=self.value) Number to convert
        @type  bit_count: Integer
        @param bit_count: (Optional, def=self.width) Width of bit string

        @rtype:  String
        @return: Bit string
        '''
        if number is None:
            if isinstance(self.value, (list, tuple)):
                # We have been given a list to cycle through that is not being mutated...
                if self.cyclic_index == len(self.value):
                    # Reset the index.
                    self.cyclic_index = 0
                number = self.value[self.cyclic_index]
                self.cyclic_index += 1
            else:
                number = self.value

        if bit_count is None:
            bit_count = self.width

        return "".join(map(lambda x: str((number >> x) & 1), range(bit_count - 1, -1, -1)))

    def to_decimal(self, binary):
        '''
        Convert a binary string to a decimal number.

        @type  binary: String
        @param binary: Binary string

        @rtype:  Integer
        @return: Converted bit string
        '''

        return int(binary, 2)


########################################################################################################################
class byte (bit_field):
    def __init__(self, value, endian="<", format="binary", signed=False, full_range=False, fuzzable=True, name=None):
        self.s_type = "byte"
        if not isinstance(value, (int, list, tuple)):
            value = struct.unpack(endian + "B", value)[0]

        bit_field.__init__(self, value, 8, None, endian, format, signed, full_range, fuzzable, name)


########################################################################################################################
class word (bit_field):
    def __init__(self, value, endian="<", format="binary", signed=False, full_range=False, fuzzable=True, name=None):
        self.s_type = "word"
        if not isinstance(value, (int, list, tuple)):
            value = struct.unpack(endian + "H", value)[0]

        bit_field.__init__(self, value, 16, None, endian, format, signed, full_range, fuzzable, name)


########################################################################################################################
class dword (bit_field):
    def __init__(self, value, endian="<", format="binary", signed=False, full_range=False, fuzzable=True, name=None):
        self.s_type = "dword"
        if not isinstance(value, (int, list, tuple)):
            value = struct.unpack(endian + "L", value)[0]

        bit_field.__init__(self, value, 32, None, endian, format, signed, full_range, fuzzable, name)


########################################################################################################################
class qword (bit_field):
    def __init__(self, value, endian="<", format="binary", signed=False, full_range=False, fuzzable=True, name=None):
        self.s_type = "qword"
        if not isinstance(value, (int, list, tuple)):
            value = struct.unpack(endian + "Q", value)[0]

        bit_field.__init__(self, value, 64, None, endian, format, signed, full_range, fuzzable, name)



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2023, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com

import copy
from chipsec.parsers import BaseConfigParser
from chipsec.parsers import Stage
from chipsec.parsers import info_data

CONFIG_TAG = 'configuration'


def _get_range_data(xml_node, attr):
    int_items = []
    for item in xml_node.attrib[attr].split(','):
        item = item.strip()
        if item.upper().endswith('*'):
            x = int(item.replace('*', '0'), 0)
            int_items.extend(range(x, x + 0x10))
        elif '-' in item:
            item_min, item_max = item.split('-', 1)
            int_items.extend(range(int(item_min, 0), int(item_max, 0) + 1))
        else:
            int_items.append(int(item, 0))
    return int_items


def _config_convert_data(xml_node, did_is_range=False):
    INT_KEYS = ['dev', 'fun', 'vid', 'did', 'rid', 'offset',
                'bit', 'size', 'port', 'msr', 'value', 'address',
                'fixed_address', 'base_align', 'align_bits', 'mask',
                'reg_align', 'limit_align', 'regh_align',
                'width', 'reg']
    BOOL_KEYS = ['req_pch']
    INT_LIST_KEYS = ['bus']
    STR_LIST_KEYS = ['config']
    RANGE_LIST_KEYS = ['detection_value']
    if did_is_range:
        INT_KEYS.remove('did')
        RANGE_LIST_KEYS.append('did')
    node_data = {}
    for key in xml_node.attrib:
        if key in INT_KEYS:
            node_data[key] = int(xml_node.attrib[key], 0)
        elif key in INT_LIST_KEYS:
            node_data[key] = [int(xml_node.attrib[key], 0)]
        elif key in STR_LIST_KEYS:
            node_data[key] = [x.strip() for x in xml_node.attrib[key].split(',')]
        elif key in RANGE_LIST_KEYS:
            node_data[key] = _get_range_data(xml_node, key)
        elif key in BOOL_KEYS:
            node_data[key] = xml_node.attrib[key].lower() == 'true'
        else:
            node_data[key] = xml_node.attrib[key]
    return node_data


class PlatformInfo(BaseConfigParser):
    def get_metadata(self):
        return {'info': self.handle_info}

    def get_stage(self):
        return Stage.GET_INFO

    def handle_info(self, et_node, stage_data):
        platform = ''
        req_pch = None
        family = None
        proc_code = None
        pch_code = None
        detect_vals = []
        sku_data = []
        vid_int = int(stage_data.vid_str, 16)

        # Extract platform information. If no platform found it is just a device entry.
        cfg_info = _config_convert_data(stage_data.configuration)
        if 'platform' in cfg_info:
            platform = cfg_info['platform']
        if 'req_pch' in cfg_info:
            req_pch = cfg_info['req_pch']
        if platform and platform.lower().startswith('pch'):
            pch_code = platform.upper()
        else:
            proc_code = platform.upper()

        # Start processing the <info> tag
        for info in et_node.iter('info'):
            cfg_info = _config_convert_data(info)
            if 'family' in cfg_info:
                family = cfg_info['family']
            if 'detection_value' in cfg_info:
                detect_vals = cfg_info['detection_value']
            for sku in info.iter('sku'):
                sku_info = _config_convert_data(sku, True)
                if 'code' not in sku_info or sku_info['code'] != platform.upper():
                    sku_info['code'] = platform.upper()
                if 'vid' not in sku_info:
                    sku_info['vid'] = vid_int
                sku_data.append(sku_info)

        return info_data(family, proc_code, pch_code, detect_vals, req_pch, stage_data.vid_str, sku_data)


class CoreConfig(BaseConfigParser):
    def get_metadata(self):
        return {'pci': self.handle_pci,
                'mmio': self.handle_mmio,
                'io': self.handle_io,
                'ima': self.handle_ima,
                'memory': self.handle_memory,
                'registers': self.handle_registers,
                'controls': self.handle_controls,
                'locks': self.handle_locks}

    def get_stage(self):
        return Stage.DEVICE_CFG

    def _process_pci_dev(self, vid_str, dev_name, dev_attr):
        device_added = False
        if 'did' in dev_attr:
            for did in dev_attr['did']:
                did_str = self.cfg._make_hex_key_str(did)
                if vid_str in self.cfg.CONFIG_PCI_RAW and did_str in self.cfg.CONFIG_PCI_RAW[vid_str]:
                    pci_data = self.cfg.CONFIG_PCI_RAW[vid_str][did_str]
                    self._add_dev(vid_str, dev_name, pci_data, dev_attr)
                    device_added = True
                    break
        else:
            if vid_str in self.cfg.CONFIG_PCI_RAW:
                for did_str in self.cfg.CONFIG_PCI_RAW[vid_str]:
                    pci_data = self.cfg.CONFIG_PCI_RAW[vid_str][did_str]

                    if dev_attr['bus'] in pci_data['bus'] and dev_attr['dev'] == pci_data['dev'] and \
                    dev_attr['fun'] == pci_data['fun']:
                        self._add_dev(vid_str, dev_name, pci_data, dev_attr)
                        device_added = True
                        break
        if not device_added:
            self._add_dev(vid_str, dev_name, None, dev_attr)

    def _add_dev(self, vid_str, name, pci_info, dev_attr):
        if pci_info:
            self.cfg.BUS[name] = pci_info['bus']
            self.cfg.CONFIG_PCI[name] = copy.copy(pci_info)
        else:
            self.cfg.CONFIG_PCI[name] = copy.deepcopy(dev_attr)
            self.cfg.BUS[name] = []
            if 'did' in dev_attr:
                self.cfg.CONFIG_PCI[name]['did'] = dev_attr['did'][0]

    def handle_pci(self, et_node, stage_data):
        ret_val = []

        for dev in et_node.iter('device'):
            dev_attr = _config_convert_data(dev, True)
            if 'name' not in dev_attr:
                continue
            dev_name = dev_attr['name']
            self._process_pci_dev(stage_data.vid_str, dev_name, dev_attr)
            self.logger.log_debug(f"    + {dev_attr['name']:16}: {dev_attr}")

        return ret_val

    def handle_controls(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.CONTROLS, stage_data, et_node, 'control')

    def handle_io(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.IO_BARS, stage_data, et_node, 'bar')

    def handle_ima(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.IMA_REGISTERS, stage_data, et_node, 'indirect')

    def handle_locks(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.LOCKS, stage_data, et_node, 'lock')

    def handle_memory(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.MEMORY_RANGES, stage_data, et_node, 'range')

    def handle_mmio(self, et_node, stage_data):
        return self._add_entry_simple(self.cfg.MMIO_BARS, stage_data, et_node, 'bar')

    def handle_registers(self, et_node, stage_data):
        ret_val = []
        dest = self.cfg.REGISTERS
        for reg in et_node.iter('register'):
            reg_attr = _config_convert_data(reg)
            if 'name' not in reg_attr:
                self.logger.log_error(f'Missing name entry for {reg_attr}')
                continue
            reg_name = reg_attr['name']
            if 'undef' in reg_attr:
                if reg_name in dest:
                    self.logger.log_debug(f"    - {reg_name:16}: {reg_attr['undef']}")
                    dest.pop(reg_name, None)
                continue

            # Patch missing or incorrect data
            if 'desc' not in reg_attr:
                reg_attr['desc'] = reg_name
            if 'size' not in reg_attr:
                self.logger.log_debug(f'Missing size entry for {reg_name:16}: {reg_attr}. Assuming 4 bytes')
                reg_attr['size'] = 4

            # Get existing field data
            if reg_name in self.cfg.REGISTERS:
                reg_fields = self.cfg.REGISTERS[reg_name]['FIELDS']
            else:
                reg_fields = {}

            for field in reg.iter('field'):
                field_attr = _config_convert_data(field)
                field_name = field_attr['name']

                # Locked by attributes need to be handled here due to embedding information in field data
                if 'lockedby' in field_attr:
                    lockedby = field_attr['lockedby']
                    if lockedby in self.cfg.LOCKEDBY:
                        self.cfg.LOCKEDBY[lockedby].append({reg_name, field_name})
                    else:
                        self.cfg.LOCKEDBY[lockedby] = [{reg_name, field_name}]
                # Handle rest of field data here
                if 'desc' not in field_attr:
                    field_attr['desc'] = field_name
                reg_fields[field_name] = field_attr

            # Store all register data
            reg_attr['FIELDS'] = reg_fields
            self.cfg.REGISTERS[reg_name] = reg_attr
            self.logger.log_debug(f'    + {reg_name:16}: {reg_attr}')
        return ret_val

    def _add_entry_simple(self, dest, stage_data, et_node, node_name):
        ret_val = []
        for node in et_node.iter(node_name):
            attrs = _config_convert_data(node)
            if 'name' not in attrs:
                self.logger.log_error(f'Missing name entry for {attrs}')
                continue
            if 'undef' in attrs:
                if attrs['name'] in dest:
                    self.logger.log_debug(f"    - {attrs['name']:16}: {attrs['undef']}")
                    dest.pop(attrs['name'], None)
                continue
            if 'desc' not in attrs:
                attrs['desc'] = attrs['name']
            dest[attrs['name']] = attrs
            self.logger.log_debug(f"    + {attrs['name']:16}: {attrs}")
        return ret_val


parsers = [PlatformInfo, CoreConfig]



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util vmm hypercall <rax> <rbx> <rcx> <rdx> <rdi> <rsi> [r8] [r9] [r10] [r11]
>>> chipsec_util vmm hypercall <eax> <ebx> <ecx> <edx> <edi> <esi>
>>> chipsec_util vmm pt|ept <ept_pointer>
>>> chipsec_util vmm virtio [<bus>:<device>.<function>]

Examples:

>>> chipsec_util vmm hypercall 32 0 0 0 0 0
>>> chipsec_util vmm pt 0x524B01E
>>> chipsec_util vmm virtio
>>> chipsec_util vmm virtio 0:6.0
"""

import re

from chipsec.command import BaseCommand, toLoad
from chipsec.hal.vmm import VMM, get_virtio_devices, VirtIO_Device
from chipsec.hal.pci import print_pci_devices
from chipsec.library.exceptions import VMMRuntimeError
from argparse import ArgumentParser


class VMMCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.Driver

    def parse_arguments(self) -> None:
        parser = ArgumentParser(prog='chipsec_util vmm', usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_hypercall = subparsers.add_parser('hypercall')
        parser_hypercall.add_argument('ax', type=lambda x: int(x, 16), help='rax/eax value (hex)')
        parser_hypercall.add_argument('bx', type=lambda x: int(x, 16), help='rbx/ebx value (hex)')
        parser_hypercall.add_argument('cx', type=lambda x: int(x, 16), help='rcx/ecx value (hex)')
        parser_hypercall.add_argument('dx', type=lambda x: int(x, 16), help='rdx/edx value (hex)')
        parser_hypercall.add_argument('di', type=lambda x: int(x, 16), help='rdi/edi value (hex)')
        parser_hypercall.add_argument('si', type=lambda x: int(x, 16), help='rsi/esi value (hex)')
        parser_hypercall.add_argument('r8', type=lambda x: int(x, 16), nargs='?', default=0, help='r8 value (hex)')
        parser_hypercall.add_argument('r9', type=lambda x: int(x, 16), nargs='?', default=0, help='r9 value (hex)')
        parser_hypercall.add_argument('r10', type=lambda x: int(x, 16), nargs='?', default=0, help='r10 value (hex)')
        parser_hypercall.add_argument('r11', type=lambda x: int(x, 16), nargs='?', default=0, help='r11 value (hex)')
        parser_hypercall.set_defaults(func=self.vmm_hypercall)

        parser_pt = subparsers.add_parser('pt')
        parser_pt.add_argument('eptp', type=lambda x: int(x, 16), help='Pointer (hex)')
        parser_pt.set_defaults(func=self.vmm_pt)

        parser_ept = subparsers.add_parser('ept')
        parser_ept.add_argument('eptp', type=lambda x: int(x, 16), help='Pointer (hex)')
        parser_ept.set_defaults(func=self.vmm_pt)

        parser_virtio = subparsers.add_parser('virtio')
        parser_virtio.add_argument('bdf', type=str, nargs='?', default=None, help='<bus>:<device>.<function>')
        parser_virtio.set_defaults(func=self.vmm_virtio)

        parser.parse_args(self.argv, namespace=self)

    def vmm_virtio(self):
        if self.bdf is not None:
            match = re.search(r"^([0-9a-f]{1,2}):([0-1]?[0-9a-f]{1})\.([0-7]{1})$", self.bdf)
            if match:
                _bus = int(match.group(1), 16) & 0xFF
                _dev = int(match.group(2), 16) & 0x1F
                _fun = int(match.group(3), 16) & 0x07
                vid = self.cs.pci.read_word(_bus, _dev, _fun, 0)
                did = self.cs.pci.read_word(_bus, _dev, _fun, 2)
                dev = (_bus, _dev, _fun, vid, did)
                virt_dev = [dev]
            else:
                self.logger.log_error("Invalid B:D.F ({})".format(self.bdf))
                self.logger.log(VMMCommand.__doc__)
                return
        else:
            self.logger.log("[CHIPSEC] Enumerating VirtIO devices...")
            virt_dev = get_virtio_devices(self.cs.pci.enumerate_devices())

        if len(virt_dev) > 0:
            self.logger.log("[CHIPSEC] Available VirtIO devices:")
            print_pci_devices(virt_dev)
            for (b, d, f, vid, did, rid) in virt_dev:
                VirtIO_Device(self.cs, b, d, f).dump_device()
        else:
            self.logger.log("[CHIPSEC] No VirtIO devices found")

    def vmm_hypercall(self):
        self.logger.log('')
        self.logger.log("[CHIPSEC] > hypercall")
        self.logger.log("[CHIPSEC]   RAX: 0x{:016X}".format(self.ax))
        self.logger.log("[CHIPSEC]   RBX: 0x{:016X}".format(self.bx))
        self.logger.log("[CHIPSEC]   RCX: 0x{:016X}".format(self.cx))
        self.logger.log("[CHIPSEC]   RDX: 0x{:016X}".format(self.dx))
        self.logger.log("[CHIPSEC]   RSI: 0x{:016X}".format(self.si))
        self.logger.log("[CHIPSEC]   RDI: 0x{:016X}".format(self.di))
        self.logger.log("[CHIPSEC]   R8 : 0x{:016X}".format(self.r8))
        self.logger.log("[CHIPSEC]   R9 : 0x{:016X}".format(self.r9))
        self.logger.log("[CHIPSEC]   R10: 0x{:016X}".format(self.r10))
        self.logger.log("[CHIPSEC]   R11: 0x{:016X}".format(self.r11))

        rax = self.vmm.hypercall(self.ax, self.bx, self.cx, self.dx, self.si, self.di, self.r8, self.r9, self.r10, self.r11)

        self.logger.log("[CHIPSEC] < RAX: 0x{:016X}".format(rax))

    def vmm_pt(self):
        if self.eptp is not None:
            pt_fname = 'ept_{:08X}'.format(self.eptp)
            self.logger.log("[CHIPSEC] EPT physical base: 0x{:016X}".format(self.eptp))
            self.logger.log("[CHIPSEC] Dumping EPT to '{}'...".format(pt_fname))
            self.vmm.dump_EPT_page_tables(self.eptp, pt_fname)
        else:
            self.logger.log("[CHIPSEC] Finding EPT hierarchy in memory is not implemented yet")
            self.logger.log_error(VMMCommand.__doc__)
            return

    def run(self):
        try:
            self.vmm = VMM(self.cs)
        except VMMRuntimeError as msg:
            self.logger.log_error(msg)
            return

        self.vmm.init()

        self.func()


commands = {'vmm': VMMCommand}



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

"""
>>> chipsec_util spd detect
>>> chipsec_util spd dump [device_addr]
>>> chipsec_util spd read <device_addr> <offset>
>>> chipsec_util spd write <device_addr> <offset> <byte_val>

Examples:

>>> chipsec_util spd detect
>>> chipsec_util spd dump DIMM0
>>> chipsec_util spd dump 0xA0
>>> chipsec_util spd read DIMM2 0x0
>>> chipsec_util spd read 0xA0 0x0
>>> chipsec_util spd write 0xA0 0x0 0xAA
"""

from chipsec.command import BaseCommand, toLoad
from chipsec.hal import smbus, spd
from argparse import ArgumentParser


class SPDCommand(BaseCommand):

    def requirements(self) -> toLoad:
        return toLoad.All

    def parse_arguments(self) -> None:
        parser = ArgumentParser(usage=__doc__)
        subparsers = parser.add_subparsers()

        parser_detect = subparsers.add_parser('detect')
        parser_detect.set_defaults(func=self.spd_detect)

        parser_dump = subparsers.add_parser('dump')
        parser_dump.add_argument('dev', type=str, nargs='?', default=None, help="Device")
        parser_dump.set_defaults(func=self.spd_dump)

        parser_read = subparsers.add_parser('read')
        parser_read.add_argument('dev', type=str, help="Device Address")
        parser_read.add_argument('off', type=lambda x: int(x, 16), nargs='?', default=None, help="Offset (hex)")
        parser_read.set_defaults(func=self.spd_read)

        parser_write = subparsers.add_parser('write')
        parser_write.add_argument('dev', type=str, help="Device Address")
        parser_write.add_argument('off', type=lambda x: int(x, 16), help="Offset (hex)")
        parser_write.add_argument('val', type=lambda x: int(x, 16), help="Byte Value (hex)")
        parser_write.set_defaults(func=self.spd_write)

        parser.parse_args(self.argv, namespace=self)

    def spd_detect(self):
        self.logger.log("[CHIPSEC] Searching for DIMMs with SPD...")
        _dimms = self._spd.detect()
        if _dimms is not None:
            self.logger.log("Detected the following SPD devices:")
            for _dimm in _dimms:
                self.logger.log("{}: 0x{:02X}".format(spd.SPD_DIMMS[_dimm], _dimm))
        else:
            self.logger.log("Unable to detect SPD devices.")

    def spd_dump(self):
        if self.dev is not None:
            _dev = self.dev.upper()
            self.dev_addr = spd.SPD_DIMM_ADDRESSES[_dev] if _dev in spd.SPD_DIMM_ADDRESSES else int(self.dev, 16)
            if not self._spd.isSPDPresent(self.dev_addr):
                self.logger.log("[CHIPSEC] SPD for DIMM 0x{:X} is not found".format(self.dev_addr))
                return
            self._spd.decode(self.dev_addr)
        else:
            _dimms = self._spd.detect()
            for _dimm in _dimms:
                self._spd.decode(_dimm)

    def spd_read(self):
        _dev = self.dev.upper()
        self.dev_addr = spd.SPD_DIMM_ADDRESSES[_dev] if _dev in spd.SPD_DIMM_ADDRESSES else int(self.dev, 16)
        if not self._spd.isSPDPresent(self.dev_addr):
            self.logger.log("[CHIPSEC] SPD for DIMM 0x{:X} is not found".format(self.dev_addr))
            return

        val = self._spd.read_byte(self.off, self.dev_addr)
        self.logger.log("[CHIPSEC] SPD read: offset 0x{:X} = 0x{:X}".format(self.off, val))

    def spd_write(self):
        _dev = self.dev.upper()
        self.dev_addr = spd.SPD_DIMM_ADDRESSES[_dev] if _dev in spd.SPD_DIMM_ADDRESSES else int(self.dev, 16)
        if not self._spd.isSPDPresent(self.dev_addr):
            self.logger.log("[CHIPSEC] SPD for DIMM 0x{:X} is not found".format(self.dev_addr))
            return

        self.logger.log("[CHIPSEC] SPD write: offset 0x{:X} = 0x{:X}".format(self.off, self.val))
        self._spd.write_byte(self.off, self.val, self.dev_addr)

    def run(self):
        try:
            _smbus = smbus.SMBus(self.cs)
            self._spd = spd.SPD(_smbus)
        except BaseException as msg:
            self.logger.log_error(msg)
            return
        if not _smbus.is_SMBus_supported():
            self.logger.log("[CHIPSEC] SMBus controller is not supported")
            return
        self.dev_addr = spd.SPD_SMBUS_ADDRESS
        self.func()


commands = {'spd': SPDCommand}



# CHIPSEC: Platform Security Assessment Framework
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
#

import unittest

from tests.software import mock_helper, util


class TestCMOSChipsecUtil(util.TestChipsecUtil):
    """Test the CMOS commands exposed by chipsec_utils."""

    def test_cmos_dump(self):
        """Test to verify the output of 'cmos dump'.

        Check that we only access CMOS IO ports.
        """

        class CMOSHelper(mock_helper.TestHelper):
            def read_io_port(self, io_port, size):
                if io_port < 0x70 or io_port > 0x73:
                    raise Exception("Reading outside CMOS IO port")
                return io_port

            def write_io_port(self, io_port, value, size):
                if io_port < 0x70 or io_port > 0x73:
                    raise Exception("Writing outside CMOS IO port")

        self._chipsec_util("cmos dump", CMOSHelper)


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2ExitCode.OK23, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  ExitCode.OK211ExitCode.OK-13ExitCode.OK1, USA.
#
# Contact information:
# chipsec@intel.com
#

# To execute: python -m unittest tests.modules.test_tgl_modules

import unittest
import os

from chipsec.library.file import get_main_dir
from chipsec.testcase import ExitCode
from tests.modules.run_chipsec_module import setup_run_destroy_module_with_mock_logger


class TestTglModules(unittest.TestCase):
    def setUp(self) -> None:
        self.folder_path = os.path.join(get_main_dir(), "tests", "modules", "tgl")
        self.init_replay_file = os.path.join(self.folder_path, "enumeration.json")

    def derive_filename(self, module_name: str) -> str:
        return f"{module_name.replace('.', '-')}_test.json"

    def run_and_test_module(self, module_name: str, expected_returncode: int) -> None:
        test_recording = self.derive_filename(module_name)
        replay_file = os.path.join(self.folder_path, test_recording)
        retval = setup_run_destroy_module_with_mock_logger(self.init_replay_file, module_name, module_replay_file=replay_file)
        self.assertEqual(retval, expected_returncode, f"Expected: {expected_returncode} but got: {retval}")

    def test_tgl_module_bios_smi(self):
        self.run_and_test_module("common.bios_smi", ExitCode.OK)

    def test_tgl_module_bios_ts(self):
        self.run_and_test_module("common.bios_ts", ExitCode.OK)

    def test_tgl_module_bios_wp(self):
        self.run_and_test_module("common.bios_wp", ExitCode.OK)

    def test_tgl_module_cpu_cpu_info(self):
        self.run_and_test_module("common.cpu.cpu_info", ExitCode.INFORMATION)

    def test_tgl_module_cpu_ia_untrusted(self):
        self.run_and_test_module("common.cpu.ia_untrusted", ExitCode.OK)

    def test_tgl_module_cpu_spectre_v2(self):
        self.run_and_test_module("common.cpu.spectre_v2", ExitCode.OK)

    def test_tgl_module_debugenabled(self):
        self.run_and_test_module("common.debugenabled", ExitCode.OK)

    def test_tgl_module_ia32cfg(self):
        self.run_and_test_module("common.ia32cfg", ExitCode.OK)

    def test_tgl_module_memconfig(self):
        self.run_and_test_module("common.memconfig", ExitCode.OK)

    def test_tgl_module_memlock(self):
        self.run_and_test_module("common.memlock", ExitCode.NOTAPPLICABLE)

    def test_tgl_module_me_mfg_mode(self):
        self.run_and_test_module("common.me_mfg_mode", ExitCode.OK)

    def test_tgl_module_remap(self):
        self.run_and_test_module("common.remap", ExitCode.OK)

    def test_tgl_module_rtclock(self):
        self.run_and_test_module("common.rtclock", ExitCode.WARNING)

    def test_tgl_module_secureboot_variables(self):
        self.run_and_test_module("common.secureboot.variables", ExitCode.OK)

    def test_tgl_module_sgx_check(self):
        self.run_and_test_module("common.sgx_check", ExitCode.NOTAPPLICABLE)

    def test_tgl_module_smm_code_chk(self):
        self.run_and_test_module("common.smm_code_chk", ExitCode.OK)

    def test_tgl_module_smm_dma(self):
        self.run_and_test_module("common.smm_dma", ExitCode.OK)

    def test_tgl_module_smm(self):
        self.run_and_test_module("common.smm", ExitCode.NOTAPPLICABLE)

    def test_tgl_module_smrr(self):
        self.run_and_test_module("common.smrr", ExitCode.OK)

    def test_tgl_module_spd_wd(self):
        self.run_and_test_module("common.spd_wd", ExitCode.OK)

    def test_tgl_module_spi_access(self):
        self.run_and_test_module("common.spi_access", ExitCode.FAIL)

    def test_tgl_module_spi_desc(self):
        self.run_and_test_module("common.spi_desc", ExitCode.OK)

    def test_tgl_module_spi_fdopss(self):
        self.run_and_test_module("common.spi_fdopss", ExitCode.OK)

    def test_tgl_module_spi_lock(self):
        self.run_and_test_module("common.spi_lock", ExitCode.OK)

    def test_tgl_module_uefi_access_uefispec(self):
        self.run_and_test_module("common.uefi.access_uefispec", ExitCode.OK)

    @unittest.skip("S3bootscript module was archived")
    def test_tgl_module_uefi_s3bootscript(self):
        self.run_and_test_module("common.uefi.s3bootscript", ExitCode.WARNING)


if __name__ == '__main__':
    unittest.main()



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2010-2022, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

import struct
import os
import string
import platform
from collections.abc import Iterable
from typing import Any, Dict, Tuple, Optional, AnyStr
from chipsec.library.file import get_main_dir

BIT0 = 0x0001
BIT1 = 0x0002
BIT2 = 0x0004
BIT3 = 0x0008
BIT4 = 0x0010
BIT5 = 0x0020
BIT6 = 0x0040
BIT7 = 0x0080
BIT8 = 0x0100
BIT9 = 0x0200
BIT10 = 0x0400
BIT11 = 0x0800
BIT12 = 0x1000
BIT13 = 0x2000
BIT14 = 0x4000
BIT15 = 0x8000
BIT16 = 0x00010000
BIT17 = 0x00020000
BIT18 = 0x00040000
BIT19 = 0x00080000
BIT20 = 0x00100000
BIT21 = 0x00200000
BIT22 = 0x00400000
BIT23 = 0x00800000
BIT24 = 0x01000000
BIT25 = 0x02000000
BIT26 = 0x04000000
BIT27 = 0x08000000
BIT28 = 0x10000000
BIT29 = 0x20000000
BIT30 = 0x40000000
BIT31 = 0x80000000
BIT32 = 0x100000000
BIT33 = 0x200000000
BIT34 = 0x400000000
BIT35 = 0x800000000
BIT36 = 0x1000000000
BIT37 = 0x2000000000
BIT38 = 0x4000000000
BIT39 = 0x8000000000
BIT40 = 0x10000000000
BIT41 = 0x20000000000
BIT42 = 0x40000000000
BIT43 = 0x80000000000
BIT44 = 0x100000000000
BIT45 = 0x200000000000
BIT46 = 0x400000000000
BIT47 = 0x800000000000
BIT48 = 0x1000000000000
BIT49 = 0x2000000000000
BIT50 = 0x4000000000000
BIT51 = 0x8000000000000
BIT52 = 0x10000000000000
BIT53 = 0x20000000000000
BIT54 = 0x40000000000000
BIT55 = 0x80000000000000
BIT56 = 0x100000000000000
BIT57 = 0x200000000000000
BIT58 = 0x400000000000000
BIT59 = 0x800000000000000
BIT60 = 0x1000000000000000
BIT61 = 0x2000000000000000
BIT62 = 0x4000000000000000
BIT63 = 0x8000000000000000

BOUNDARY_1KB = 0x400
BOUNDARY_2KB = 0x800
BOUNDARY_4KB = 0x1000
BOUNDARY_1MB = 0x100000
BOUNDARY_2MB = 0x200000
BOUNDARY_4MB = 0x400000
BOUNDARY_8MB = 0x800000
BOUNDARY_16MB = 0x1000000
BOUNDARY_32MB = 0x2000000
BOUNDARY_64MB = 0x4000000
BOUNDARY_128MB = 0x8000000
BOUNDARY_256MB = 0x10000000
BOUNDARY_512MB = 0x20000000
BOUNDARY_1GB = 0x40000000
BOUNDARY_2GB = 0x80000000
BOUNDARY_4GB = 0x100000000

ALIGNED_4KB = 0xFFF
ALIGNED_1MB = 0xFFFFF
ALIGNED_8MB = 0x7FFFFF
ALIGNED_64MB = 0x3FFFFFF
ALIGNED_128MB = 0x7FFFFFF
ALIGNED_256MB = 0xFFFFFFF

MASK_8b = 0xFF
MASK_16b = 0xFFFF
MASK_32b = 0xFFFFFFFF
MASK_64b = 0xFFFFFFFFFFFFFFFF

CHIPSET_CODE_UNKNOWN = ''
PCH_CODE_PREFIX = 'PCH_'
CHIPSET_ID_UNKNOWN = 0


def bit(bit_num: int) -> int:
    return int(1 << bit_num)


def is_set(val: int, bit_mask: int) -> bool:
    return bool(val & bit_mask != 0)


def scan_single_bit_mask(bit_mask: int) -> Optional[int]:
    for bit_num in range(0, 7):
        if bit_mask >> bit_num == 1:
            return bit_num
    return None


def DB(val: int) -> bytes:
    return struct.pack('<B', val)


def DW(val: int) -> bytes:
    return struct.pack('<H', val)


def DD(val: int) -> bytes:
    return struct.pack('<L', val)


def DQ(val: int) -> bytes:
    return struct.pack('<Q', val)


SIZE2FORMAT: Dict[int, str] = {
    1: 'B',
    2: 'H',
    4: 'I',
    8: 'Q'
}


def bytestostring(mbytes: AnyStr) -> str:
    if isinstance(mbytes, bytes) or isinstance(mbytes, bytearray):
        return mbytes.decode("latin_1")
    else:
        return mbytes


def stringtobytes(mstr: AnyStr) -> bytes:
    if isinstance(mstr, str):
        return mstr.encode("latin_1")
    else:
        return mstr


def pack1(value: int, size: int) -> bytes:
    """Shortcut to pack a single value into a string based on its size."""
    return struct.pack(SIZE2FORMAT[size], value)


def unpack1(string: bytes, size: int) -> int:
    """Shortcut to unpack a single value from a string based on its size."""
    return struct.unpack(SIZE2FORMAT[size], string)[0]


def get_bits(value: int, start: int, nbits: int) -> int:
    ret = value >> start
    ret &= (1 << nbits) - 1
    return ret


def get_version() -> str:
    version_strs = []
    chipsec_folder = os.path.abspath(get_main_dir())
    for fname in sorted([x for x in os.listdir(os.path.join(chipsec_folder, "chipsec")) if x.startswith('VERSION')]):
        version_file = os.path.join(chipsec_folder, "chipsec", fname)
        with open(version_file, "r") as verFile:
            version_strs.append(verFile.read().strip())
    return '-'.join(version_strs)


def os_version() -> Tuple[str, str, str, str]:
    return platform.system(), platform.release(), platform.version(), platform.machine()


def is_printable(seq: AnyStr) -> bool:
    return set(bytestostring(seq)).issubset(set(string.printable))


def is_hex(maybe_hex: Iterable) -> bool:
    return all(char in string.hexdigits for char in maybe_hex)


def is_all_ones(value: int, size: int, width: int = 8) -> bool:
    mask = (1 << (size * width)) - 1
    return (mask == (mask & value))


def get_message() -> str:
    msg_str = ""
    chipsec_folder = os.path.abspath(get_main_dir())
    msg_file = os.path.join(chipsec_folder, "chipsec", "MESSAGE")
    if os.path.exists(msg_file):
        with open(msg_file, "r") as msgFile:
            msg_str = msgFile.read()
    return msg_str


def is_all_value(value_list: list, value: Any) -> bool:
    """Checks if all elements in a list are equal to a given value"""
    return all(n == value for n in value_list)


class ARCH_VID:
    INTEL = 0x8086
    AMD = 0x1022



# CHIPSEC: Platform Security Assessment Framework
# Copyright (c) 2019-2021, Intel Corporation
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; Version 2.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# Contact information:
# chipsec@intel.com
#

from abc import ABC, abstractmethod
from typing import Dict, List, Tuple, Optional, TYPE_CHECKING
if TYPE_CHECKING:
    from chipsec.library.types import EfiVariableType
    from ctypes import Array

# Base class for the helpers


class Helper(ABC):
    class __metaclass__(type):
        def __init__(cls, name, bases, attrs):
            if not hasattr(cls, 'registry'):
                cls.registry = []
            else:
                cls.registry.append((name, cls))

    @abstractmethod
    def __init__(self):
        self.driver_loaded = False
        self.os_system = 'basehelper'
        self.os_release = '0.0'
        self.os_version = '0.0'
        self.os_machine = 'base'
        self.name = 'Helper'
        self.driverpath = ''

    @abstractmethod
    def create(self) -> bool:
        pass

    @abstractmethod
    def start(self) -> bool:
        pass

    @abstractmethod
    def stop(self) -> bool:
        pass

    @abstractmethod
    def delete(self) -> bool:
        pass

    def get_info(self) -> Tuple[str, str]:
        return self.name, self.driverpath

    #################################################################################################
    # Actual OS helper functionality accessible to HAL components

    #
    # Read/Write PCI configuration registers via legacy CF8/CFC ports
    #
    @abstractmethod
    def read_pci_reg(self, bus: int, device: int, function: int, address: int, size: int) -> int:
        pass

    @abstractmethod
    def write_pci_reg(self, bus: int, device: int, function: int, address: int, value: int, size: int) -> int:
        pass

    #
    # read/write mmio
    #
    @abstractmethod
    def read_mmio_reg(self, phys_address: int, size: int) -> int:
        pass

    @abstractmethod
    def write_mmio_reg(self, phys_address: int, size: int, value: int) -> int:
        pass

    #
    # physical_address is 64 bit integer
    #
    @abstractmethod
    def read_phys_mem(self, phys_address: int, size: int) -> bytes:
        pass

    @abstractmethod
    def write_phys_mem(self, phys_address: int, size: int, buffer: bytes) -> int:
        pass

    @abstractmethod
    def alloc_phys_mem(self, size: int, max_phys_address: int) -> Tuple[int, int]:
        pass

    @abstractmethod
    def free_phys_mem(self, phys_address: int):
        pass

    @abstractmethod
    def va2pa(self, virtual_address: int) -> Tuple[int, int]:
        pass

    @abstractmethod
    def map_io_space(self, phys_address: int, size: int, cache_type: int) -> int:
        pass

    #
    # Read/Write I/O port
    #
    @abstractmethod
    def read_io_port(self, io_port: int, size: int) -> int:
        pass

    @abstractmethod
    def write_io_port(self, io_port: int, value: int, size: int) -> int:
        pass

    #
    # Read/Write CR registers
    #
    @abstractmethod
    def read_cr(self, cpu_thread_id: int, cr_number: int) -> int:
        pass

    @abstractmethod
    def write_cr(self, cpu_thread_id: int, cr_number: int, value: int) -> int:
        pass

    #
    # Read/Write MSR on a specific CPU thread
    #
    @abstractmethod
    def read_msr(self, cpu_thread_id: int, msr_addr: int) -> Tuple[int, int]:
        pass

    @abstractmethod
    def write_msr(self, cpu_thread_id: int, msr_addr: int, eax: int, edx: int) -> int:
        pass

    #
    # Load CPU microcode update on a specific CPU thread
    #
    @abstractmethod
    def load_ucode_update(self, cpu_thread_id: int, ucode_update_buffer: bytes) -> bool:
        pass

    #
    # Read IDTR/GDTR/LDTR on a specific CPU thread
    #
    @abstractmethod
    def get_descriptor_table(self, cpu_thread_id: int, desc_table_code: int) -> Optional[Tuple[int, int, int]]:
        pass

    #
    # EFI Variable API
    #
    @abstractmethod
    def EFI_supported(self) -> bool:
        pass

    @abstractmethod
    def get_EFI_variable(self, name: str, guid: str) -> Optional[bytes]:
        pass

    @abstractmethod
    def set_EFI_variable(self, name: str, guid: str, buffer: bytes, buffer_size: Optional[int], attrs: Optional[int]) -> Optional[int]:
        pass

    @abstractmethod
    def delete_EFI_variable(self, name: str, guid: str) -> Optional[int]:
        pass

    @abstractmethod
    def list_EFI_variables(self) -> Optional[Dict[str, List['EfiVariableType']]]:
        pass

    #
    # ACPI
    #
    @abstractmethod
    def enum_ACPI_tables(self) -> Optional['Array']:
        pass

    @abstractmethod
    def get_ACPI_table(self, table_name: str) -> Optional['Array']:
        pass

    #
    # CPUID
    #
    @abstractmethod
    def cpuid(self, eax: int, ecx: int) -> Tuple[int, int, int, int]:
        pass

    #
    # IOSF Message Bus access
    #
    @abstractmethod
    def msgbus_send_read_message(self, mcr: int, mcrx: int) -> Optional[int]:
        pass

    @abstractmethod
    def msgbus_send_write_message(self, mcr: int, mcrx: int, mdr: int) -> None:
        pass

    @abstractmethod
    def msgbus_send_message(self, mcr: int, mcrx: int, mdr: Optional[int]) -> Optional[int]:
        pass

    #
    # Affinity
    #
    @abstractmethod
    def get_affinity(self) -> Optional[int]:
        pass

    @abstractmethod
    def set_affinity(self, value: int) -> Optional[int]:
        pass

    #
    # Logical CPU count
    #
    @abstractmethod
    def get_threads_count(self) -> int:
        pass

    #
    # Send SW SMI
    #
    @abstractmethod
    def send_sw_smi(self, cpu_thread_id: int, SMI_code_data: int, _rax: int, _rbx: int, _rcx: int, _rdx: int, _rsi: int, _rdi: int) -> Optional[int]:
        pass

    #
    # Hypercall
    #
    @abstractmethod
    def hypercall(self, rcx: int, rdx: int, r8: int, r9: int, r10: int, r11: int, rax: int, rbx: int, rdi: int, rsi: int, xmm_buffer: int) -> int:
        pass

    #
    # Speculation control
    #
    @abstractmethod
    def retpoline_enabled(self) -> bool:
        pass
